#!/usr/bin/env python
# License: BSD 3 clause
"""
Loads a trained model and outputs predictions based on input feature files.

:author: Nitin Madnani (nmadnani@ets.org)
:author: Dan Blanchard (dblanchard@ets.org)
:organization: ETS
:date: February 2013
"""

import argparse
import logging
import sys
from pathlib import Path
from typing import List, Optional, Union

import numpy as np

from skll.data.readers import EXT_TO_READER
from skll.learner import Learner
from skll.types import LabelType
from skll.version import __version__


def main(argv: Optional[List[str]] = None):
    """
    Handle command line arguments and gets things started.

    Parameters
    ----------
    argv : Optional[List[str]], default=None
        List of arguments, as if specified on the command-line.
        If ``None``, ``sys.argv[1:]`` is used instead.
    """
    # Get command line arguments
    parser = argparse.ArgumentParser(
        description="Loads a trained model and outputs predictions based "
        "on input feature files.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        conflict_handler="resolve",
    )
    parser.add_argument(
        "model_file", help="Model file to load and use for generating " "predictions."
    )
    parser.add_argument(
        "input_files",
        help="A space-separated list of CSV, TSV, or "
        "jsonlines files (with or without the label "
        "column), with the appropriate suffix.",
        nargs="+",
    )
    parser.add_argument(
        "-i",
        "--id_col",
        help="Name of the column which contains the instance " "IDs in ARFF, CSV, or TSV files.",
        default="id",
    )
    parser.add_argument(
        "-l",
        "--label_col",
        help="Name of the column which contains the labels "
        "in ARFF, CSV, or TSV files. For ARFF files, "
        "this must be the final column to count as the "
        "label.",
        default="y",
    )
    group = parser.add_mutually_exclusive_group()
    group.add_argument(
        "-p",
        "--predict_labels",
        help="If the model is doing probabilistic "
        "classification, output the class label with the "
        "highest probability instead of the class "
        "probabilities.",
        action="store_true",
        default=False,
    )
    group.add_argument(
        "-t",
        "--threshold",
        help="If the model we're using is doing probabilistic "
        "binary classification, output the positive class"
        " label if its probability meets/exceeds this "
        "threshold and output the negative class label "
        "otherwise.",
        type=float,
    )
    parser.add_argument(
        "-q", "--quiet", help='Suppress printing of "Loading..." messages.', action="store_true"
    )
    parser.add_argument(
        "-o",
        "--output_file",
        help="Path to output tsv file. If not specified, "
        "predictions will be printed to stdout. For "
        "probabilistic binary classification, the "
        "probability of the positive class will always "
        "be in the last column.",
    )
    parser.add_argument("--version", action="version", version=f"%(prog)s {__version__}")

    args = parser.parse_args(argv)

    # Make warnings from built-in warnings module get formatted more nicely
    logging.captureWarnings(True)
    logging.basicConfig(format=("%(asctime)s - %(name)s - %(levelname)s - " "%(message)s"))
    logger = logging.getLogger(__name__)

    # load the model from disk
    learner = Learner.from_file(args.model_file)

    # is the model a regressor or a classifier?
    estimator_type = learner.model_type._estimator_type

    # if we are using a binary classification model, get the positive
    # class label string if from the `pos_label` attribute, and if
    # that is `None`, get it from the learner's label dictionary; also
    # get the string denoting the negative label which is just the other
    # label in the list
    if estimator_type == "classifier":
        if len(learner.label_list) == 2:
            if learner.pos_label is not None:
                pos_label = learner.pos_label
            else:
                pos_label = [
                    label for label in learner.label_dict if learner.label_dict[label] == 1
                ][0]
            neg_label = [label for label in learner.label_list if label != pos_label][0]
            logger.info(f"{pos_label} is the label for the positive " "class.")

    # if we want to choose labels by thresholding the probabilities,
    # make sure that the learner is probabilistic AND binary first
    is_probabilistic_classifier = hasattr(learner._model, "predict_proba") and learner.probability
    if args.threshold is not None and (
        not is_probabilistic_classifier or len(learner.label_list) != 2
    ):
        error_msg = (
            "Cannot threshold probabilities to predict positive "
            f"class since given {learner._model_type.__name__} "
            "learner is either multi-class, non-probabilistic, or "
            "was not trained with probability=True."
        )
        logger.error(error_msg)
        raise ValueError(error_msg)

    # if we want to choose labels by predicting the most likely label,
    # make sure that the learner is probabilistic
    if args.predict_labels and not is_probabilistic_classifier:
        error_msg = (
            "Cannot predict most likely labels from probabilities "
            f"since given {learner._model_type.__name__} learner is "
            "either non-probabilistic or was not trained with "
            "probability=True."
        )
        logger.error(error_msg)
        raise ValueError(error_msg)

    # iterate over all the specified input files
    for i, input_file in enumerate(args.input_files):
        # make sure each file extension is one we can process
        input_extension = Path(input_file).suffix.lower()
        if input_extension not in EXT_TO_READER:
            logger.error(
                f"Input file must be in either .arff, .csv, "
                f".jsonlines, .libsvm, .ndj, or .tsv format. "
                f" Skipping file {input_file}"
            )
            continue
        else:
            # read in the file into a featureset
            reader = EXT_TO_READER[input_extension](
                input_file, quiet=args.quiet, label_col=args.label_col, id_col=args.id_col
            )
            feature_set = reader.read()

            # for this featureset, get the predictions of either the
            # most likely class labels or the class label probabilities;
            # if the model is a regressor then `class_labels` will be
            # ignored entirely
            original_predictions = learner.predict(
                feature_set, class_labels=not learner.probability or args.predict_labels
            )

            # get the appropriate header depending on the what we will
            # be outputting; if we are using a regressor or a non-probabilistic
            # learner, or thresholding probabilities, or predicting most likely
            # labels, we are outputting only two columns - the ID and the label,
            # otherwise we are outputting N + 1 columns where N = number of classes
            if (
                estimator_type == "regressor"
                or not learner.probability
                or args.predict_labels
                or args.threshold is not None
            ):
                header = ["id", "prediction"]
            else:
                header = ["id"] + [str(x) for x in learner.label_list]

            # now let us start computing what we want to output based
            # on the predictions we have so far

            # Threshold the positive class label probability
            predictions: Union[np.ndarray, List[LabelType]]
            if args.threshold is not None:
                predictions = []
                for _, pos_label_prob in original_predictions:
                    chosen_label = pos_label if pos_label_prob >= args.threshold else neg_label
                    predictions.append(chosen_label)

            # For everything else, we can just use the original predictions
            else:
                predictions = original_predictions

            # now initialize the output file
            try:
                outputfh = open(args.output_file, "a") if args.output_file else sys.stdout

                # write out the header first but only once
                if i == 0:
                    print("\t".join(header), file=outputfh)

                # and now write out the predictions
                for j, prediction in enumerate(predictions):
                    id_ = feature_set.ids[j]
                    if isinstance(prediction, (np.ndarray, list)):
                        prediction_str = "\t".join([str(p) for p in prediction])
                    else:
                        prediction_str = prediction
                    print(f"{id_}\t{prediction_str}", file=outputfh)

            finally:
                # close the file if we had opened one
                if args.output_file:
                    outputfh.close()


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# License: BSD 3 clause
"""
Runs scikit-learn jobs given a config file.

:author: Dan Blanchard (dblanchard@ets.org)
:author: Michael Heilman (mheilman@ets.org)
:author: Nitin Madnani (nmadnani@ets.org)
"""


import logging
from argparse import ArgumentDefaultsHelpFormatter, ArgumentParser
from typing import List, Optional

from skll.experiments import run_configuration
from skll.version import __version__


def main(argv: Optional[List[str]] = None) -> None:
    """
    Handle command line arguments and get things started.

    Parameters
    ----------
    argv : Optional[List[str]], default=None
        List of arguments, as if specified on the command-line.
        If ``None``, ``sys.argv[1:]`` is used instead.
    """
    # Get command line arguments
    parser = ArgumentParser(
        description="Runs the scikit-learn experiments "
        "in a given config file. If Grid Map "
        "is installed, jobs will automatically"
        " be created and run on a "
        "DRMAA-compatible cluster.",
        formatter_class=ArgumentDefaultsHelpFormatter,
        conflict_handler="resolve",
    )
    parser.add_argument(
        "config_file", help="Configuration file describing the task to run.", nargs="+"
    )
    parser.add_argument(
        "-a",
        "--ablation",
        help="Runs an ablation study where repeated "
        "experiments are conducted where the specified "
        "number of features in each featureset in the "
        "configuration file are held out.",
        type=int,
        default=0,
        metavar="NUM_FEATURES",
    )
    parser.add_argument(
        "-A",
        "--ablation_all",
        help="Runs an ablation study where repeated "
        "experiments are conducted with all combinations "
        "of features in each featureset in the "
        "configuration file. Overrides --ablation "
        "setting.",
        action="store_true",
    )
    parser.add_argument(
        "-k",
        "--keep_models",
        help="If trained models already exists, re-use them " "instead of overwriting them.",
        action="store_true",
    )
    parser.add_argument(
        "-l",
        "--local",
        help="Do not use the Grid Engine for running jobs and"
        " just run everything sequentially on the local "
        "machine. ",
        action="store_true",
    )
    parser.add_argument(
        "-m",
        "--machines",
        help="comma-separated list of machines to add to the "
        "gridmap whitelist (if not specified, all "
        "available machines are used). Note that full "
        "names must be specified, e.g., "
        '"nlp.research.ets.org"',
        default=None,
    )
    parser.add_argument("-q", "--queue", help="Use this queue for gridmap.", default="all.q")
    parser.add_argument(
        "-r",
        "--resume",
        help="If result files already exist for an experiment"
        " do not overwrite them. This is very useful "
        "when doing a large ablation experiment and part"
        " of it crashes.",
        action="store_true",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        help="Include debug information in the logging output.",
        default=False,
        action="store_true",
    )
    parser.add_argument("--version", action="version", version=f"%(prog)s {__version__}")
    args = parser.parse_args(argv)

    # Default logging level is INFO unless we are being verbose
    log_level = logging.DEBUG if args.verbose else logging.INFO

    # Make warnings from built-in warnings module get formatted more nicely
    logging.captureWarnings(True)
    logging.basicConfig(
        format="%(asctime)s - %(name)s - %(levelname)s - " "%(message)s", level=log_level
    )

    machines = None
    if args.machines:
        machines = args.machines.split(",")

    ablation = args.ablation
    if args.ablation_all:
        ablation = None

    # Run each config file sequentially
    for config_file in args.config_file:
        run_configuration(
            config_file,
            local=args.local,
            overwrite=not args.keep_models,
            queue=args.queue,
            hosts=machines,
            ablation=ablation,
            resume=args.resume,
            log_level=log_level,
        )


if __name__ == "__main__":
    main()



"""
Utility classes and functions for SKLL learners.

:author: Nitin Madnani (nmadnani@ets.org)
:author: Michael Heilman (mheilman@ets.org)
:author: Dan Blanchard (dblanchard@ets.org)
:author: Aoife Cahill (acahill@ets.org)
:organization: ETS
"""

from __future__ import annotations

import inspect
import logging
import sys
import time
from collections import Counter, defaultdict
from csv import DictWriter, excel_tab
from functools import wraps
from importlib import import_module
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Iterable,
    List,
    Optional,
    Set,
    Tuple,
    Type,
    Union,
)

import joblib
import numpy as np
import scipy.sparse as sp
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.feature_selection import SelectKBest
from sklearn.metrics import (
    accuracy_score,
    confusion_matrix,
    precision_recall_fscore_support,
)
from sklearn.model_selection import (
    KFold,
    LeaveOneGroupOut,
    ShuffleSplit,
    StratifiedKFold,
)

from skll.data import FeatureSet
from skll.metrics import _CUSTOM_METRICS, _PREDEFINED_CUSTOM_METRICS, use_score_func
from skll.types import (
    ComputeEvalMetricsResults,
    ConfusionMatrix,
    FeaturesetIterator,
    FoldMapping,
    IdType,
    IndexIterator,
    LabelType,
    PathOrStr,
    SparseFeatureMatrix,
)
from skll.utils.constants import (
    CLASSIFICATION_ONLY_METRICS,
    CORRELATION_METRICS,
    REGRESSION_ONLY_METRICS,
    UNWEIGHTED_KAPPA_METRICS,
    WEIGHTED_KAPPA_METRICS,
)
from skll.version import VERSION

# import classes that we only need for type checking and not
# otherwise since they would cause circular import issues
if TYPE_CHECKING:
    import skll.learner
    import skll.learner.voting


class Densifier(BaseEstimator, TransformerMixin):
    """
    Custom pipeline stage for handling dense feature arrays.

    A custom pipeline stage that will be inserted into the
    learner pipeline attribute to accommodate the situation
    when SKLL needs to manually convert feature arrays from
    sparse to dense. For example, when features are being hashed
    but we are also doing centering using the feature means.
    """

    def fit(self, X, y=None):
        """Fit the estimator."""
        return self

    def fit_transform(self, X, y=None):
        """Fit the estimator and transform the input."""
        return self

    def transform(self, X):
        """Transform the input using already fit estimator."""
        return X.toarray()


class FilteredLeaveOneGroupOut(LeaveOneGroupOut):
    """
    Custom version ``LeaveOneGroupOut`` cross-validation iterator.

    This version only outputs indices of instances with IDs in a prespecified set.

    Parameters
    ----------
    keep : Iterable[IdType]
        A set of IDs to keep.
    example_ids : numpy.ndarray, of length n_samples
        A list of example IDs.
    logger : Optional[logging.Logger], default=None
        A logger instance.
    """

    def __init__(
        self,
        keep: Iterable[IdType],
        example_ids: np.ndarray,
        logger: Optional[logging.Logger] = None,
    ):
        """Initialize the model."""
        super(FilteredLeaveOneGroupOut, self).__init__()
        self.keep = keep
        self.example_ids = example_ids
        self._warned = False
        self.logger = logger if logger else logging.getLogger(__name__)

    def split(
        self, X: SparseFeatureMatrix, y: np.ndarray, groups: Optional[List[str]]
    ) -> IndexIterator:
        """
        Generate indices to split data into training and test set.

        Parameters
        ----------
        X : numpy.ndarray, with shape (n_samples, n_features)
            Training data, where n_samples is the number of samples
            and n_features is the number of features.
        y : numpy.ndarray, of length n_samples
            The target variable for supervised learning problems.
        groups : List[str]
            Group labels for the samples used while splitting the dataset into
            train/test set.

        Yields
        ------
        train_index : numpy.ndarray
            The training set indices for that split.
        test_index : numpy.ndarray
            The testing set indices for that split.
        """
        for train_index, test_index in super(FilteredLeaveOneGroupOut, self).split(X, y, groups):
            train_len = len(train_index)
            test_len = len(test_index)
            train_index = [i for i in train_index if self.example_ids[i] in self.keep]
            test_index = [i for i in test_index if self.example_ids[i] in self.keep]
            if not self._warned and (train_len != len(train_index) or test_len != len(test_index)):
                self.logger.warning(
                    "Feature set contains IDs that are not "
                    "in folds dictionary. Skipping those IDs."
                )
                self._warned = True

            yield train_index, test_index


class SelectByMinCount(SelectKBest):
    """
    Select or discard features based on how often they occur in the data.

    Select features occurring in more (and/or fewer than) than a specified
    number of examples in the training data (or a CV training fold).

    Parameters
    ----------
    min_count : int, default=1
        The minimum feature count to select.
    """

    def __init__(self, min_count: int = 1):
        """Initialize the model."""
        self.min_count = min_count
        self.scores_: Optional[np.ndarray] = None

    def fit(self, X, y=None):
        """
        Fit the SelectByMinCount model.

        Parameters
        ----------
        X : numpy.ndarray, with shape (n_samples, n_features)
            The training data to fit.
        y : Ignored
            Not used.

        Returns
        -------
        self
        """
        # initialize a list of counts of times each feature appears
        col_counts = [0 for _ in range(X.shape[1])]

        if sp.issparse(X):
            # find() is scipy.sparse's equivalent of nonzero()
            _, col_indices, _ = sp.find(X)
        else:
            # assume it's a numpy array (not a numpy matrix)
            col_indices = X.nonzero()[1].tolist()

        for i in col_indices:
            col_counts[i] += 1

        self.scores_ = np.array(col_counts)
        return self

    def _get_support_mask(self):
        """
        Return a mask indicating which features to keep.

        Adapted from ``SelectKBest``.

        Returns
        -------
        mask : numpy.ndarray
            The mask with features to keep set to True.
        """
        mask = np.zeros(self.scores_.shape, dtype=bool)
        mask[self.scores_ >= self.min_count] = True
        return mask


def add_unseen_labels(
    train_label_dict: Dict[LabelType, int], test_label_list: List[LabelType]
) -> Dict[LabelType, int]:
    """
    Merge test set labels that are not seen in the training data with seen ones.

    Parameters
    ----------
    train_label_dict : Dict[:class:`skll.types.LabelType`, int]
        Dictionary mapping training set class labels to class indices.
    test_label_list : List[:class:`skll.types.LabelType`]
        List containing labels in the test set.

    Returns
    -------
    Dict[:class:`skll.types.LabelType`, int]
        Dictionary mapping merged labels from both the training and test sets
        to indices.
    """
    # get the list of labels that were in the training set
    train_label_list = list(train_label_dict.keys())

    # identify any unseen labels in the test set
    unseen_test_label_list = [label for label in test_label_list if label not in train_label_list]

    # create a new dictionary for these unseen labels with label indices
    # for them starting _after_ those for the training set labels
    unseen_label_dict = {
        label: i for i, label in enumerate(unseen_test_label_list, start=len(train_label_list))
    }

    # combine the train label dictionary with this unseen label one & return
    train_and_test_label_dict = train_label_dict.copy()
    train_and_test_label_dict.update(unseen_label_dict)
    return train_and_test_label_dict


def compute_evaluation_metrics(
    metrics: List[str],
    labels: np.ndarray,
    predictions: np.ndarray,
    model_type: str,
    label_dict: Optional[Dict[LabelType, int]] = None,
    grid_objective: Optional[str] = None,
    probability: bool = False,
    logger: Optional[logging.Logger] = None,
) -> ComputeEvalMetricsResults:
    """
    Compute given evaluation metrics.

    Compute given metrics to evaluate the given predictions generated
    by the given type of estimator against the given true labels.

    Parameters
    ----------
    metrics : List[str]
        List of metrics to compute.
    labels : numpy.ndarray
        True labels to be used for computing the metrics.
    predictions : numpy.ndarray
        The predictions to be used for computing the metrics.
    model_type : str
        One of "classifier" or "regressor".
    label_dict : Optional[Dict[LabelType, int]], default=None
        Dictionary mapping class labels to indices for classification.
    grid_objective : Optional[str], default=None
        The objective used for tuning the hyper-parameters of the model
        that generated the predictions. If ``None``, it means that no
        grid search was done.
    probability : bool, default=False
        Does the model output class probabilities?
    logger : Optional[logging.Logger], default=None
        A logger instance to use for logging messages and warnings.
        If ``None``, a new one is created.

    Returns
    -------
    :class:`skll.types.ComputeEvalMetricsResults`
        5-tuple including the confusion matrix, the overall accuracy, the
        per-label PRFs, the grid search objective function score, and the
        additional evaluation metrics, if any. For regressors, the
        first two elements are ``None``.
    """
    # set up the logger
    logger = logger if logger else logging.getLogger(__name__)

    # warn if grid objective was also specified in metrics
    if len(metrics) > 0 and grid_objective in metrics:
        logger.warning(
            f"The grid objective '{grid_objective}' is also "
            "specified as an evaluation metric. Since its value "
            "is already included in the results as the objective "
            "score, it will not be printed again in the list of "
            "metrics."
        )
        metrics = [metric for metric in metrics if metric != grid_objective]

    # initialize a dictionary that will hold all of the metric scores
    metric_scores: Dict[str, Optional[float]] = {metric: None for metric in metrics}

    # if we are doing classification and are a probablistic
    # learner or a soft-voting meta learner, then `yhat` are
    # probabilities so we need to compute the class indices
    # separately and save them too
    if model_type == "classifier" and probability:
        class_probs = predictions
        predictions = np.argmax(class_probs, axis=1)
    # if we are a regressor or classifier not in probability
    # mode, then we have the class indices already and there
    # are no probabilities
    else:
        class_probs = None

    # make a single list of metrics including the grid objective
    # since it's easier to compute everything together
    metrics_to_compute = [grid_objective] + metrics
    for metric in metrics_to_compute:
        # skip the None if we are not doing grid search
        if not metric:
            continue

        # declare types
        preds_for_metric: Optional[np.ndarray]

        # CASE 1: in probability mode for classification which means we
        # need to either use the probabilities directly or infer the labels
        # from them depending on the metric
        if probability and label_dict and class_probs is not None:
            # there are three possible cases here:
            # (a) if we are using a correlation metric or
            #     `average_precision` or `roc_auc` in a binary
            #      classification scenario, then we need to explicitly
            #     pass in the probabilities of the positive class.
            # (b) if we are using `neg_log_loss`, then we
            #     just pass in the full probability array
            # (c) we compute the most likely labels from the
            #     probabilities via argmax and use those
            #     for all other metrics
            if (
                len(label_dict) == 2
                and (metric in CORRELATION_METRICS or metric in ["average_precision", "roc_auc"])
                and metric != grid_objective
            ):
                logger.info(
                    "using probabilities for the positive class to "
                    f"compute '{metric}' for evaluation."
                )
                preds_for_metric = class_probs[:, 1]
            elif metric == "neg_log_loss":
                preds_for_metric = class_probs
            else:
                preds_for_metric = predictions

        # CASE 2: no probability mode for classifier or regressor
        # in which case we just use the predictions as they are
        else:
            preds_for_metric = predictions

        try:
            metric_scores[metric] = use_score_func(metric, labels, preds_for_metric)
        except ValueError:
            metric_scores[metric] = float("NaN")

    # now separate out the grid objective score from the additional metric scores
    # if a grid objective was actually passed in. If no objective was passed in
    # then that score should just be none.
    objective_score = None
    additional_scores = metric_scores.copy()
    if grid_objective:
        objective_score = metric_scores[grid_objective]
        del additional_scores[grid_objective]

    # declare the type for the results
    res: ComputeEvalMetricsResults

    # compute some basic statistics for regressors
    if model_type == "regressor":
        regressor_result_dict: Dict[LabelType, Any] = {"descriptive": defaultdict(dict)}
        for table_label, y in zip(["actual", "predicted"], [labels, predictions]):
            regressor_result_dict["descriptive"][table_label]["min"] = min(y)
            regressor_result_dict["descriptive"][table_label]["max"] = max(y)
            regressor_result_dict["descriptive"][table_label]["avg"] = np.mean(y)
            regressor_result_dict["descriptive"][table_label]["std"] = np.std(y)
        regressor_result_dict["pearson"] = use_score_func("pearson", labels, predictions)
        res = (None, None, regressor_result_dict, objective_score, additional_scores)
    elif label_dict:
        # compute the confusion matrix and precision/recall/f1
        # note that we are using the class indices here
        # and not the actual class labels themselves
        num_labels = len(label_dict)
        conf_mat: ConfusionMatrix = confusion_matrix(
            labels, predictions, labels=list(range(num_labels))
        ).tolist()
        # Calculate metrics
        overall_accuracy: float = accuracy_score(labels, predictions)
        result_matrix = precision_recall_fscore_support(
            labels, predictions, labels=list(range(num_labels)), average=None
        )

        # Store results
        classifier_result_dict: Dict[LabelType, Any] = defaultdict(dict)
        for actual_label in sorted(label_dict):
            col = label_dict[actual_label]
            classifier_result_dict[actual_label]["Precision"] = result_matrix[0][col]
            classifier_result_dict[actual_label]["Recall"] = result_matrix[1][col]
            classifier_result_dict[actual_label]["F-measure"] = result_matrix[2][col]

        res = (
            conf_mat,
            overall_accuracy,
            classifier_result_dict,
            objective_score,
            additional_scores,
        )

    return res


def compute_num_folds_from_example_counts(
    cv_folds: int,
    labels: Optional[np.ndarray],
    model_type: str,
    logger: Optional[logging.Logger] = None,
) -> int:
    """
    Calculate number of cross-validation folds, based on number of examples per label.

    Parameters
    ----------
    cv_folds : int
        The number of cross-validation folds.
    labels : numpy.ndarray
        The example labels.
    model_type : str
        One of "classifier" or "regressor".
    logger : Optional[logging.Logger], default=None
        A logger instance to use for logging messages and warnings.
        If ``None``, a new one is created.

    Returns
    -------
    int
        The number of folds to use, based on the number of examples
        for each label.

    Raises
    ------
    ValueError
        If ``cv_folds`` is not an integer or if the training set has
        fewer than 2 examples associated with a label (for classification).
    """
    # get a logger if not provided
    logger = logger if logger else logging.getLogger(__name__)

    try:
        assert isinstance(cv_folds, int)
    except AssertionError:
        raise ValueError("`cv_folds` must be an integer.")

    # For regression models, we can just return the current cv_folds
    if model_type == "regressor":
        return cv_folds

    min_examples_per_label = min(Counter(labels).values())
    if min_examples_per_label <= 1:
        raise ValueError(
            f"The training set has only {min_examples_per_label}" " example for a label."
        )
    if min_examples_per_label < cv_folds:
        logger.warning(
            "The minimum number of examples per label was "
            f"{min_examples_per_label}. Setting the number of "
            "cross-validation folds to that value."
        )
        cv_folds = min_examples_per_label
    return cv_folds


def contiguous_ints_or_floats(numbers: np.ndarray) -> bool:
    """
    Check for continuity in the given list of numbers.

    Check whether the given list of numbers contains
    contiguous integers or contiguous integer-like
    floats. For example, [1, 2, 3] or [4.0, 5.0, 6.0]
    are both contiguous but [1.1, 1.2, 1.3] is not.

    Parameters
    ----------
    numbers : numpy.ndarray
        The numbers we want to check.

    Returns
    -------
    bool
        ``True`` if the numbers are contiguous integers
        or contiguous integer-like floats (1.0, 2.0, etc.),
        ``False`` otherwise.

    Raises
    ------
    TypeError
        If ``numbers`` does not contain integers or floating point values.
    ValueError
        If ``numbers`` is empty.
    """
    try:
        # make sure that number is not empty
        assert len(numbers) > 0

        # first check that the numbers are all integers
        # or integer-like floats (e.g., 1.0, 2.0 etc.)
        ints_or_int_like_floats = np.all(np.mod(numbers, 1) == 0)

        # next check that the successive differences between
        # the numbers are all 1, i.e., they are nuermicontiguous
        contiguous = np.all(np.diff(numbers) == 1)

    except AssertionError:
        raise ValueError("Input cannot be empty.")

    except TypeError:
        raise TypeError("Input should only contain numbers.")

    # we need both conditions to be true and we want to return
    # a regular Python `bool`, not a `numpy.bool_`
    return bool(ints_or_int_like_floats and contiguous)


def get_acceptable_classification_metrics(label_array: np.ndarray) -> Set[str]:
    """
    Return acceptable metrics given the unique set of labels being classified.

    Parameters
    ----------
    label_array : numpy.ndarray
        A sorted numpy array containing the unique labels
        that we are trying to predict. Optional for regressors
        but required for classifiers.

    Returns
    -------
    acceptable_metrics : Set[str]
        A set of metric names that are acceptable
        for the given classification scenario.
    """
    # this is a classifier so the acceptable objective
    # functions definitely include those metrics that
    # are specifically for classification and also
    # the unweighted kappa metrics
    acceptable_metrics = CLASSIFICATION_ONLY_METRICS | UNWEIGHTED_KAPPA_METRICS

    # now let us consider which other metrics may also
    # be acceptable depending on whether the labels
    # are strings or (contiguous) integers/floats
    label_type = label_array.dtype.type

    # CASE 1: labels are strings, then no other metrics
    # are acceptable
    if issubclass(label_type, (np.object_, str)):
        pass

    # CASE 2: labels are integers or floats; the way
    # it works in SKLL, it's guaranteed that
    # class indices will be sorted in the same order
    # as the class labels therefore, ranking metrics
    # such as various correlations should work fine.
    elif issubclass(label_type, (int, np.int32, np.int64, float, np.float32, np.float64)):
        acceptable_metrics.update(CORRELATION_METRICS)

        # CASE 3: labels are numerically contiguous integers
        # this is a special sub-case of CASE 2 which
        # represents ordinal classification. Only in this
        # case, weighted kappas -- where the distance
        # between the class labels has a special
        # meaning -- can be allowed. This is because
        # class indices are always contiguous and all
        # metrics in SKLL are computed in the index
        # space, not the label space. Note that floating
        # point numbers that are equivalent to integers
        # (e.g., [1.0, 2.0, 3.0]) are also acceptable.
        if contiguous_ints_or_floats(label_array):
            acceptable_metrics.update(WEIGHTED_KAPPA_METRICS)

    # if there are any user-defined custom metrics registered, include them too
    user_defined_metrics = set(_CUSTOM_METRICS) - set(_PREDEFINED_CUSTOM_METRICS)
    if len(user_defined_metrics) > 0:
        acceptable_metrics.update(user_defined_metrics)

    return acceptable_metrics


def get_acceptable_regression_metrics() -> Set[str]:
    """Return the set of metrics that are acceptable for regression."""
    # it's fairly straightforward for regression since
    # we do not have to check the labels
    acceptable_metrics = (
        REGRESSION_ONLY_METRICS
        | UNWEIGHTED_KAPPA_METRICS
        | WEIGHTED_KAPPA_METRICS
        | CORRELATION_METRICS
    )

    # if there are any user-defined custom metrics registered, include them too
    user_defined_metrics = set(_CUSTOM_METRICS) - set(_PREDEFINED_CUSTOM_METRICS)
    if len(user_defined_metrics) > 0:
        acceptable_metrics.update(user_defined_metrics)

    return acceptable_metrics


def load_custom_learner(
    custom_learner_path: Optional[PathOrStr], custom_learner_name: str
) -> "skll.learner.Learner":
    """
    Import and load the custom learner object from the given path.

    Parameters
    ----------
    custom_learner_path : :class:`skll.types.PathOrStr`
        The path to a custom learner.
    custom_learner_name : str
        The name of a custom learner.

    Returns
    -------
    :class:`skll.learner.Learner`
        The SKLL learner object loaded from the given path.

    Raises
    ------
    ValueError
        If the custom learner path does not end in '.py'.
    """
    if not custom_learner_path:
        raise ValueError(
            "custom_learner_path was not set and learner " f"{custom_learner_name} was not found."
        )

    # convert to a Path object
    custom_learner_path = Path(custom_learner_path)

    if custom_learner_path.suffix != ".py":
        raise ValueError("custom_learner_path must end in .py " f"({custom_learner_path})")

    custom_learner_module_name = custom_learner_path.stem
    sys.path.append(str(custom_learner_path.resolve().parent))
    import_module(custom_learner_module_name)
    return getattr(sys.modules[custom_learner_module_name], custom_learner_name)


def get_predictions(
    learner: Union["skll.learner.Learner", "skll.learner.voting.VotingLearner"], xtest: np.ndarray
) -> Dict[str, Any]:
    """
    Get predictions from the given learner (or meta-learner) for given features.

    The various types of predictions include:

    - "raw" predictions which are self-explanatory for regressors; for
      classifiers, these are the indices of the class labels, not the labels
      themselves.
    - "labels": for classifiers, these are the class labels; for regressors
      they are not applicable and represented as ``None``.
    - "probabilities": for classifiers, these are the class probabilities; for
      non-probabilistic classifiers or regressors, they are not applicable and
      represented as ``None``.

    Parameters
    ----------
    learner : Union[:class:`skll.learner.Learner`, :class:`skll.learner.voting.VotingLearner`]
        The already-trained ``Learner`` or ``VotingLearner`` instance that is
        used to generate the predictions.
    xtest : numpy.ndarray
        Numpy array of features on which the predictions are to be made.

    Returns
    -------
    prediction_dict : Dict[str, Any]
        Dictionary containing the three types of predictions as the keys
        and either ``None`` or a numpy array as the value.

    Raises
    ------
    NotImplementedError
        If the scikit-learn model does not implement ``predict_proba()`` to
        get the class probabilities.
    """
    # deferred import to avoid circular dependencies
    from skll.learner.voting import VotingLearner

    # initialize the prediction dictionary
    prediction_dict: Dict[str, Any] = {"raw": None, "labels": None, "probabilities": None}

    # first get the raw predictions from the underlying scikit-learn model
    # this works for both classifiers and regressors
    yhat = learner.model.predict(xtest)
    prediction_dict["raw"] = yhat

    # next, if it's a classifier ...
    if learner.model_type._estimator_type == "classifier":
        # get the list of labels from the learner (or meta-learner)
        if isinstance(learner, VotingLearner):
            label_list = learner.learners[0].label_list
        else:
            label_list = learner.label_list

        # get the predicted class labels
        class_labels = np.array([label_list[int(pred)] for pred in yhat])
        prediction_dict["labels"] = class_labels

        # then get the class probabilities too if the learner
        # (or meta-learner) is probabilistic
        if (hasattr(learner, "probability") and learner.probability) or (
            hasattr(learner, "voting") and learner.voting == "soft"
        ):
            try:
                yhat_probs = learner.model.predict_proba(xtest)
            except NotImplementedError as e:
                learner.logger.error(
                    f"Model type: {learner.model_type.__name__}\n" f"Model: {learner.model}\n"
                )
                raise e
            else:
                prediction_dict["probabilities"] = yhat_probs

    return prediction_dict


def rescaled(cls):
    """
    Create regressors that rescale their predictions.

    This decorator creates regressors that store a min and a max for the training
    data and make sure that predictions fall within that range.  They also store
    the means and SDs of the gold standard and the predictions on the training
    set to rescale the predictions (e.g., as in e-rater).

    Parameters
    ----------
    cls : BaseEstimator
        An estimator class to add rescaling to.

    Returns
    -------
    cls : BaseEstimator
        Modified version of estimator class with rescaled functions added.

    Raises
    ------
    ValueError
        If classifier cannot be rescaled (i.e. is not a regressor).
    """
    # If this class has already been run through the decorator, return it
    if hasattr(cls, "rescale"):
        return cls

    # Save original versions of functions to use later.
    orig_init = cls.__init__
    orig_fit = cls.fit
    orig_predict = cls.predict

    if cls._estimator_type == "classifier":
        raise ValueError("Classifiers cannot be rescaled. Only regressors " "can.")

    # Define all new versions of functions
    @wraps(cls.fit)
    def fit(self, X: np.ndarray, y=None):  # noqa: D417
        """
        Fit a model.

        Also store the mean, SD, max and min of the training set
        and the mean and SD of the predictions on the training set.

        Parameters
        ----------
        X : numpy.ndarray, with shape (n_samples, n_features)
            The data to fit.
        y : Ignored
            This is ignored.

        Returns
        -------
        self
        """
        # fit a regular regression model
        orig_fit(self, X, y=y)

        if self.constrain:
            # also record the training data min and max
            self.y_min = np.min(y)
            self.y_max = np.max(y)

        if self.rescale:
            # also record the means and SDs for the training set
            y_hat = orig_predict(self, X)
            self.yhat_mean = np.mean(y_hat)
            self.yhat_sd = np.std(y_hat)
            self.y_mean = np.mean(y)
            self.y_sd = np.std(y)

        return self

    @wraps(cls.predict)
    def predict(self, X: np.ndarray) -> np.ndarray:
        """
        Predict with regressor and rescale.

        Make predictions with the super class, and then adjust them using the
        stored min, max, means, and standard deviations.

        Parameters
        ----------
        self
            The instance itself
        X : numpy.ndarray, with shape (n_samples,)
            The data to predict.

        Returns
        -------
        numpy.ndarray
            The prediction results.
        """
        # get the unconstrained predictions
        res = orig_predict(self, X)

        if self.rescale:
            # convert the predictions to z-scores,
            # then rescale to match the training set distribution
            res = (((res - self.yhat_mean) / self.yhat_sd) * self.y_sd) + self.y_mean

        if self.constrain:
            # apply min and max constraints
            res = np.array([max(self.y_min, min(self.y_max, pred)) for pred in res])

        return res

    @classmethod
    @wraps(cls._get_param_names)
    def _get_param_names(class_x):
        """
        Get kwargs for superclass and add new kwargs.

        This is adapted from scikit-learn's ``BaseEstimator`` class.
        It gets the kwargs for the superclass's init method and adds the
        kwargs for newly added ``__init__()`` method.

        Parameters
        ----------
        class_x
            The superclass from which to retrieve param names.

        Returns
        -------
        List[str]
            A list of parameter names for the class's init method.

        Raises
        ------
        RuntimeError
            If `varargs` exist in the scikit-learn estimator.
        """
        # initialize the empty list of parameter names
        args = []

        try:
            # get signature of the original init method
            init = getattr(orig_init, "deprecated_original", orig_init)
            init_signature = inspect.signature(init)

            # get all parameters excluding 'self'
            original_parameters = [
                p
                for p in init_signature.parameters.values()
                if p.name != "self" and p.kind != p.VAR_KEYWORD
            ]

            # there should be no varargs
            for parameter in original_parameters:
                if parameter.kind == parameter.VAR_POSITIONAL:
                    raise RuntimeError(
                        "scikit-learn estimators should always specify their "
                        "parameters in the signature of their __init__ (no "
                        f"varargs). {cls} with constructor {init_signature} "
                        "doesn't follow this convention."
                    )
                else:
                    args.append(parameter.name)

        except TypeError:
            pass

        # now get the additional rescaling arguments
        rescale_args = list(inspect.signature(class_x.__init__).parameters.keys())

        # Remove 'self'
        rescale_args.remove("self")

        # add the rescaling arguments to the original arguments and sort
        args += rescale_args
        args.sort()

        return args

    @wraps(cls.__init__)
    def init(self, constrain: bool = True, rescale: bool = True, **kwargs):  # noqa: D417
        """
        Initialize things in the right order.

        Parameters
        ----------
        constrain : bool, default=True
            Whether to constrain predictions within min and max values.
        rescale : bool, default=True
            Whether to rescale prediction values using z-scores.
        kwargs : Dict[str, Any]
            Keyword arguments for base class.
        """
        # pylint: disable=W0201
        self.constrain = constrain
        self.rescale = rescale
        self.y_min = None
        self.y_max = None
        self.yhat_mean = None
        self.yhat_sd = None
        self.y_mean = None
        self.y_sd = None
        orig_init(self, **kwargs)

    # Override original functions with new ones
    cls.__init__ = init
    cls.fit = fit
    cls.predict = predict
    cls._get_param_names = _get_param_names
    cls.rescale = True

    # Return modified class
    return cls


def setup_cv_fold_iterator(
    cv_folds: Union[int, FoldMapping],
    examples: FeatureSet,
    model_type: str,
    stratified: bool = False,
    logger: Optional[logging.Logger] = None,
) -> Tuple[Union[FilteredLeaveOneGroupOut, KFold, StratifiedKFold], Optional[List[str]]]:
    """
    Set up a cross-validation fold iterator for the given ``FeatureSet``.

    Parameters
    ----------
    cv_folds : Union[int, :class:`skll.types.FoldMapping`]
        The number of folds to use for cross-validation, or
        a mapping from example IDs to folds.
    examples : :class:`skll.data.featureset.FeatureSet`
        The ``FeatureSet`` instance for which the CV iterator is to be computed.
    model_type : str
        One of "classifier" or "regressor".
    stratified : bool, default=False
        Should the cross-validation iterator be set up in a stratified fashion?
    logger : Optional[logging.Logger], default=None
        A logger instance to use for logging messages and warnings.
        If ``None``, a new one is created.

    Returns
    -------
    :class:`skll.types.StratifiedKFold`
        k-fold iterator
    Optional[List[str]]
        List of cross-validation groups
    """
    # explicitly declare the return types
    kfold: Union[FilteredLeaveOneGroupOut, KFold, StratifiedKFold]

    # Set up the cross-validation iterator.=
    if isinstance(cv_folds, int):
        cv_folds = compute_num_folds_from_example_counts(
            cv_folds, examples.labels, model_type, logger=logger
        )

        stratified = stratified and model_type == "classifier"
        if stratified:
            kfold = StratifiedKFold(n_splits=cv_folds)
            cv_groups = None
        else:
            kfold = KFold(n_splits=cv_folds)
            cv_groups = None
    # Otherwise cv_folds is a dict
    else:
        # if we have a mapping from IDs to folds, use it for the overall
        # cross-validation as well as the grid search within each
        # training fold.  Note that this means that the grid search
        # will use K-1 folds because the Kth will be the test fold for
        # the outer cross-validation.
        dummy_label = next(iter(cv_folds.values()))
        fold_groups = [cv_folds.get(curr_id, dummy_label) for curr_id in examples.ids]
        # Only retain IDs within folds if they're in cv_folds
        kfold = FilteredLeaveOneGroupOut(cv_folds, examples.ids, logger=logger)
        cv_groups = fold_groups

    return kfold, cv_groups


def setup_cv_split_iterator(
    cv_folds: Union[int, FoldMapping], examples: FeatureSet
) -> Tuple[FeaturesetIterator, int]:
    """
    Set up a cross-validation split iterator over the given ``FeatureSet``.

    Parameters
    ----------
    cv_folds : Union[int, :class:`skll.types.FoldMapping`]
        The number of folds to use for cross-validation, or
        a mapping from example IDs to folds.
    examples : :class:`skll.data.featureset.FeatureSet`
        The given featureset which is to be split.

    Returns
    -------
    :class:`skll.types.FeaturesetIterator`
        Iterator over the train/test featuresets
    int
        The maximum number of training samples available.
    """
    # seed the random number generator for replicability
    random_state = np.random.RandomState(123456789)

    # set up the cross-validation split iterator with 20% of
    # the data always reserved for testing
    cv = ShuffleSplit(n_splits=cv_folds, test_size=0.2, random_state=random_state)
    cv_iter = list(cv.split(examples.features, examples.labels, None))
    n_max_training_samples = len(cv_iter[0][0])

    # create an iterator over train/test featuresets based on the
    # cross-validation index iterator
    featureset_iter = (FeatureSet.split(examples, train, test) for train, test in cv_iter)

    return featureset_iter, n_max_training_samples


def train_and_score(
    learner: "skll.learner.Learner",
    train_examples: FeatureSet,
    test_examples: FeatureSet,
    metric: str,
) -> Tuple[float, float, float]:
    """
    Train learner, generate predictions, and evaluate predictions.

    A utility method to train a given learner instance on the given
    training examples, generate predictions on the training set itself
    and also the given test set, and score those predictions using the
    given metric. The method returns the train and test scores.

    If the learner has its ``probability`` attribute set to ``True``, it will
    produce probability values as predictions rather than class indices.
    In this case, this function will compute the argmax over the probability
    values to find the most likely class index and use that.

    Note that this method needs to be a top-level function since it is
    called from within ``joblib.Parallel()`` and, therefore, needs to be
    picklable which it would not be as an instancemethod of the ``Learner``
    class.

    Parameters
    ----------
    learner : :class:`skll.learner.Learner`
        A SKLL ``Learner`` instance.
    train_examples : :class:`skll.data.featureset.FeatureSet`
        The training examples.
    test_examples : :class:`skll.data.featureset.FeatureSet`
        The test examples.
    metric : str
        The scoring function passed to ``use_score_func()``.

    Returns
    -------
    float
        Output of the score function applied to predictions of
        ``learner`` on ``train_examples``.
    float
        Output of the score function applied to predictions of
        ``learner`` on ``test_examples``.
    float
        The time taken in seconds to fit the ``learner`` on
        ``train_examples``.
    """
    # capture the time before we train the model
    start_time = time.time()
    _ = learner.train(train_examples, grid_search=False, shuffle=False)

    # compute the time it took to train the model
    fit_time = time.time() - start_time

    # get the train and test class probabilities or indices (not labels)
    train_predictions = learner.predict(train_examples, class_labels=False)
    test_predictions = learner.predict(test_examples, class_labels=False)

    # recall that voting learners return a tuple from `predict()`
    if isinstance(train_predictions, tuple):
        train_predictions = train_predictions[0]
    if isinstance(test_predictions, tuple):
        test_predictions = test_predictions[0]

    # if we got probabilities, then we need to run argmax over them
    # to convert them into indices; this needs to handle both
    # regular learners as well as voting learners
    if (hasattr(learner, "probability") and learner.probability) or (
        hasattr(learner, "voting") and learner.voting == "soft"
    ):
        train_predictions = np.argmax(train_predictions, axis=1)
        test_predictions = np.argmax(test_predictions, axis=1)

    # now get the training and test labels and convert them to indices
    # but make sure to include any unseen labels in the test data
    if train_examples.labels is not None and test_examples.labels is not None:
        if learner.model_type._estimator_type == "classifier":
            test_label_list = np.unique(test_examples.labels).tolist()
            train_and_test_label_dict = add_unseen_labels(learner.label_dict, test_label_list)
            train_labels = np.array(
                [train_and_test_label_dict[label] for label in train_examples.labels]
            )
            test_labels = np.array(
                [train_and_test_label_dict[label] for label in test_examples.labels]
            )
        else:
            train_labels = train_examples.labels
            test_labels = test_examples.labels

    # now compute and return the scores
    train_score = use_score_func(metric, train_labels, train_predictions)
    test_score = use_score_func(metric, test_labels, test_predictions)

    return train_score, test_score, fit_time


def write_predictions(
    example_ids: np.ndarray,
    predictions_to_write: np.ndarray,
    file_prefix: str,
    model_type: str,
    label_list: List[LabelType],
    append: bool = False,
):
    """
    Write example IDs and predictions to a tab-separated file with given prefix.

    Parameters
    ----------
    example_ids : numpy.ndarray
        The IDs of the examples for which the predictions have been generated.
    predictions_to_write : numpy.ndarray
        The predictions to write out to the file.
    file_prefix : str
        The prefix for the output file. The output file will be named
        "<file_prefix>_predictions.tsv".
    model_type : str
        One of "classifier" or "regressor".
    label_list : List[:class:`skll.types.LabelType`]
        List of class labels, required if ``probability`` is ``True``.
    append : bool, default=False
        Should we append the current predictions to the file if it exists?
    """
    # create a new file starting with the given prefix
    prediction_file = f"{file_prefix}_predictions.tsv"
    with open(prediction_file, mode="w" if not append else "a", newline="") as predictionfh:
        # create a DictWriter with the appropriate field names
        if predictions_to_write.ndim > 1 and label_list:
            fieldnames = ["id"] + [label for label in label_list]
        else:
            fieldnames = ["id", "prediction"]
        writer = DictWriter(predictionfh, fieldnames=fieldnames, dialect=excel_tab)

        # write out the header unless we are appending
        if not append:
            writer.writeheader()

        # explicitly declare some types
        row: Dict[LabelType, Any]

        for example_id, pred in zip(example_ids, predictions_to_write):
            # for regressors, we just write out the prediction as-is
            if model_type == "regressor":
                row = {"id": example_id, "prediction": pred}

            # if we have an array as a prediction, it must be
            # a list of probabilities and if not, then it's
            # either a class label or an index
            else:
                if isinstance(pred, np.ndarray):
                    row = {"id": example_id}
                    row.update(dict(zip(label_list, pred)))  # type: ignore
                else:
                    row = {"id": example_id, "prediction": pred}

            # write out the row
            writer.writerow(row)


def _save_learner_to_disk(
    learner: Union["skll.learner.Learner", "skll.learner.voting.VotingLearner"], filepath: PathOrStr
) -> None:
    """
    Save the given SKLL learner instance to disk.

    NOTE: This function should only be used by the ``save()`` methods
    for the various learner classes in SKLL.

    Parameters
    ----------
    learner : Union[:class:`skll.learner.Learner`, :class:`skll.learner.voting.VotingLearner`]
        A ``Learner`` or ``VotingLearner`` instance to save to disk.
    filepath : :class:`skll.types.PathOrStr`
        The path to save the learner instance to.
    """
    # create the directory if it doesn't exist
    learner_dir = Path(filepath).parent
    if not learner_dir.exists():
        learner_dir.mkdir(parents=True)
    # write out the learner to disk
    joblib.dump((VERSION, learner), filepath)


def _load_learner_from_disk(
    learner_type: Union[Type["skll.learner.Learner"], Type["skll.learner.voting.VotingLearner"]],
    filepath: PathOrStr,
    logger: logging.Logger,
) -> Union["skll.learner.Learner", "skll.learner.voting.VotingLearner"]:
    """
    Load a saved instance of the given type from disk.

    NOTE: This function should only be used by the ``from_file()``
    methods for the various learner classes in SKLL.

    Parameters
    ----------
    learner_type : Union[Type[:class:`skll.learner.Learner`], Type[:class:`skll.learner.voting.VotingLearner`]]
        The type of learner instance to load from disk.
    filepath : :class:`skll.types.PathOrStr`
        The path to a saved ``Learner`` or ``VotingLearner`` file.
    logger : logging.Logger
        A logging object.

    Returns
    -------
    learner : Union[:class:`skll.learner.Learner`, :class:`skll.learner.voting.VotingLearner`]
        The ``Learner`` or ``VotingLearner`` instance loaded from the file.

    Raises
    ------
    ValueError
        If the pickled version of the ``Learner`` instance is out of date.
    """
    skll_version, learner = joblib.load(filepath)

    # Check that we've actually loaded an instance of the requested type
    if not isinstance(learner, learner_type):
        raise ValueError(
            f"'{filepath}' does not contain an object " f"of type '{learner_type.__name__}'."
        )

    # check that versions are compatible
    elif skll_version < (2, 5, 0):
        model_version_str = ".".join(map(str, skll_version))
        current_version_str = ".".join(map(str, VERSION))
        raise ValueError(
            f"The learner stored in '{filepath}' was "
            f"created with v{model_version_str} of SKLL, "
            "which is incompatible with the current "
            f"v{current_version_str}."
        )
    else:
        if not hasattr(learner, "sampler"):
            learner.sampler = None

        # For backward compatibility, convert string model types to actual classes
        if isinstance(learner._model_type, str):
            learner._model_type = globals()[learner._model_type]

        # set the learner logger attribute to the logger that's passed in
        learner.logger = logger

        # if the loaded learner is a `VotingLearner` then we need to attach
        # the same logger to the underlying learners as well
        if learner_type.__name__ == "VotingLearner":
            for underlying_learner in learner._learners:
                underlying_learner.logger = logger

        return learner



# License: BSD 3 clause
"""
Metrics that can be used to evaluate the performance of learners.

:author: Nitin Madnani (nmadnani@ets.org)
:author: Michael Heilman (mheilman@ets.org)
:author: Dan Blanchard (dblanchard@ets.org)
:organization: ETS
"""

import copy
import sys
from importlib import import_module
from inspect import signature
from pathlib import Path
from typing import Optional, Union

import numpy as np
from scipy.stats import kendalltau, pearsonr, spearmanr
from sklearn.metrics import (
    confusion_matrix,
    f1_score,
    fbeta_score,
    get_scorer,
    get_scorer_names,
    make_scorer,
)

from skll.types import PathOrStr


def kappa(
    y_true: np.ndarray,
    y_pred: np.ndarray,
    weights: Optional[Union[str, np.ndarray]] = None,
    allow_off_by_one: bool = False,
) -> float:
    """
    Calculate the kappa inter-rater agreement.

    The agreement is calculated between the gold standard and the predicted
    ratings. Potential values range from -1 (representing complete disagreement)
    to 1 (representing complete agreement).  A kappa value of 0 is expected if
    all agreement is due to chance.

    In the course of calculating kappa, all items in ``y_true`` and ``y_pred`` will
    first be converted to floats and then rounded to integers.

    It is assumed that y_true and y_pred contain the complete range of possible
    ratings.

    This function contains a combination of code from yorchopolis's kappa-stats
    and Ben Hamner's Metrics projects on Github.

    Parameters
    ----------
    y_true : numpy.ndarray
        The true/actual/gold labels for the data.
    y_pred : numpy.ndarray
        The predicted/observed labels for the data.
    weights : Optional[Union[str, numpy.ndarray]], default=None
        Specifies the weight matrix for the calculation.
        Possible values are: ``None`` (unweighted-kappa), ``"quadratic"``
        (quadratically weighted kappa), ``"linear"`` (linearly weighted kappa),
        and a two-dimensional numpy array (a custom matrix of weights). Each
        weight in this array corresponds to the :math:`w_{ij}` values in the
        Wikipedia description of how to calculate weighted Cohen's kappa.
    allow_off_by_one : bool, default=False
        If true, ratings that are off by one are counted as
        equal, and all other differences are reduced by
        one. For example, 1 and 2 will be considered to be
        equal, whereas 1 and 3 will have a difference of 1
        for when building the weights matrix.

    Returns
    -------
    float
        The weighted or unweighted kappa score.

    Raises
    ------
    AssertionError
        If ``y_true`` != ``y_pred``.
    ValueError
        If labels cannot be converted to int.
    ValueError
        If invalid weight scheme.
    """
    # Ensure that the lists are both the same length
    assert len(y_true) == len(y_pred)

    # This rather crazy looking typecast is intended to work as follows:
    # If an input is an int, the operations will have no effect.
    # If it is a float, it will be rounded and then converted to an int
    # because the ml_metrics package requires ints.
    # If it is a str like "1", then it will be converted to a (rounded) int.
    # If it is a str that can't be typecast, then the user is
    # given a hopefully useful error message.
    try:
        y_true = np.array([int(np.round(float(y))) for y in y_true])
        y_pred = np.array([int(np.round(float(y))) for y in y_pred])
    except ValueError:
        raise ValueError(
            "For kappa, the labels should be integers or strings"
            " that can be converted to ints (E.g., '4.0' or "
            "'3')."
        )

    # Figure out normalized expected values
    min_rating = min(min(y_true), min(y_pred))
    max_rating = max(max(y_true), max(y_pred))

    # shift the values so that the lowest value is 0
    # (to support scales that include negative values)
    y_true = y_true - min_rating
    y_pred = y_pred - min_rating

    # Build the observed/confusion matrix
    num_ratings = max_rating - min_rating + 1
    observed = confusion_matrix(y_true, y_pred, labels=list(range(num_ratings)))
    num_scored_items = float(len(y_true))

    # Build weight array if weren't passed one
    if isinstance(weights, str):
        wt_scheme = weights
        weights = None
    else:
        wt_scheme = ""

    if weights is None:
        kappa_weights = np.empty((num_ratings, num_ratings))
        for i in range(num_ratings):
            for j in range(num_ratings):
                diff = abs(i - j)
                if allow_off_by_one and diff:
                    diff -= 1
                if wt_scheme == "linear":
                    kappa_weights[i, j] = diff
                elif wt_scheme == "quadratic":
                    kappa_weights[i, j] = diff**2
                elif not wt_scheme:  # unweighted
                    kappa_weights[i, j] = bool(diff)
                else:
                    raise ValueError("Invalid weight scheme specified for " f"kappa: {wt_scheme}")
    else:
        kappa_weights = weights

    hist_true: np.ndarray = np.bincount(y_true, minlength=num_ratings)
    hist_true = hist_true[:num_ratings] / num_scored_items
    hist_pred: np.ndarray = np.bincount(y_pred, minlength=num_ratings)
    hist_pred = hist_pred[:num_ratings] / num_scored_items
    expected = np.outer(hist_true, hist_pred)

    # Normalize observed array
    observed = observed / num_scored_items

    # If all weights are zero, that means no disagreements matter.
    k = 1.0
    if np.count_nonzero(kappa_weights):
        observed_sum = np.sum(kappa_weights * observed)
        expected_sum = np.sum(kappa_weights * expected)
        k -= np.sum(observed_sum) / np.sum(expected_sum)

    return k


def correlation(y_true: np.ndarray, y_pred: np.ndarray, corr_type: str = "pearson") -> float:
    """
    Calculate given correlation type between ``y_true`` and ``y_pred``.

    ``y_pred`` can be multi-dimensional. If ``y_pred`` is 1-dimensional, it
    may either contain probabilities, most-likely classification labels, or
    regressor predictions. In that case, we simply return the correlation
    between ``y_true`` and ``y_pred``. If ``y_pred`` is multi-dimensional,
    it contains probabilties for multiple classes in which case, we infer the
    most likely labels and then compute the correlation between those and
    ``y_true``.

    Parameters
    ----------
    y_true : numpy.ndarray
        The true/actual/gold labels for the data.
    y_pred : numpy.ndarray
        The predicted/observed labels for the data.
    corr_type : str, default="pearson"
        Which type of correlation to compute. Possible
        choices are "pearson", "spearman", and "kendall_tau".

    Returns
    -------
    float
        correlation value if well-defined, else 0.0
    """
    # get the correlation function to use based on the given type
    corr_func = pearsonr
    if corr_type == "spearman":
        corr_func = spearmanr
    elif corr_type == "kendall_tau":
        corr_func = kendalltau

    # convert to numpy array in case we are passed a list
    y_pred = np.array(y_pred)

    # multi-dimensional -> probability array -> get label
    if y_pred.ndim > 1:
        labels = np.argmax(y_pred, axis=1)
        ret_score = corr_func(y_true, labels)[0]
    # 1-dimensional -> probabilities/labels -> use as is
    else:
        ret_score = corr_func(y_true, y_pred)[0]
    return ret_score


def f1_score_least_frequent(y_true: np.ndarray, y_pred: np.ndarray) -> float:
    """
    Calculate F1 score of the least frequent label/class.

    Parameters
    ----------
    y_true : numpy.ndarray
        The true/actual/gold labels for the data.
    y_pred : numpy.ndarray
        The predicted/observed labels for the data.

    Returns
    -------
    float
        F1 score of the least frequent label.
    """
    least_frequent = np.bincount(y_true).argmin()
    return f1_score(y_true, y_pred, average=None)[least_frequent]


def register_custom_metric(custom_metric_path: PathOrStr, custom_metric_name: str):
    """
    Import, load, and register the custom metric function from the given path.

    Parameters
    ----------
    custom_metric_path : :class:`skll.types.PathOrStr`
        The path to a custom metric.
    custom_metric_name : str
        The name of the custom metric function to load. This function must take
        only two array-like arguments: the true labels and the predictions,
        in that order.

    Raises
    ------
    ValueError
        If the custom metric path does not end in '.py'.
    NameError
        If the name of the custom metric file conflicts
        with an already existing attribute in ``skll.metrics``
        or if the custom metric name conflicts with a scikit-learn
        or SKLL metric.
    """
    if not custom_metric_path:
        raise ValueError(
            f"custom metric path was not set and " f"metric {custom_metric_name} was not found."
        )

    custom_metric_path = Path(custom_metric_path)
    if not custom_metric_path.exists():
        raise ValueError(f"custom metric path '{custom_metric_path}' " f"does not exist.")

    if custom_metric_path.suffix != ".py":
        raise ValueError(
            f"custom metric path must end in .py, you specified " f"{custom_metric_path}"
        )

    # get the name of the module containing the custom metric
    custom_metric_module_name = custom_metric_path.stem

    # once we know that the module name is okay, we need to make sure
    # that the metric function name is also okay
    if custom_metric_name in get_scorer_names() or custom_metric_name in _CUSTOM_METRICS:
        raise NameError(
            f"a metric called '{custom_metric_name}' already "
            f"exists; rename the metric function in "
            f"{custom_metric_module_name}.py and try again."
        )

    # dynamically import the module unless we have already done it
    if custom_metric_module_name not in sys.modules:
        sys.path.append(str(custom_metric_path.resolve().parent))
        metric_module = import_module(custom_metric_module_name)

        # this statement is only necessary so that if we end
        # up using multiprocessing parallelization backend,
        # things are serialized properly
        globals()[custom_metric_module_name] = metric_module

    # get the metric function from this imported module
    metric_func = getattr(sys.modules[custom_metric_module_name], custom_metric_name)
    # again, we need this for multiprocessing serialization
    metric_func.__module__ = f"skll.metrics.{custom_metric_module_name}"

    # extract any "special" keyword arguments from the metric function
    metric_func_parameters = signature(metric_func).parameters
    make_scorer_kwargs = {}
    for make_scorer_kwarg in ["greater_is_better", "response_method"]:
        if make_scorer_kwarg in metric_func_parameters:
            parameter = metric_func_parameters.get(make_scorer_kwarg)
            if parameter is not None:
                parameter_value = parameter.default
                make_scorer_kwargs.update({make_scorer_kwarg: parameter_value})

    # make the scorer function with the extracted keyword arguments
    _CUSTOM_METRICS[f"{custom_metric_name}"] = make_scorer(metric_func, **make_scorer_kwargs)

    return metric_func


def use_score_func(func_name: str, y_true: np.ndarray, y_pred: np.ndarray) -> float:
    """
    Call the given scoring function.

    This takes care of handling keyword arguments that were pre-specified
    when creating the scorer. This applies any sign-flipping that was
    specified by ``make_scorer()`` when the scorer was created.

    Parameters
    ----------
    func_name : str
        The name of the objective function to use.
    y_true : numpy.ndarray
        The true/actual/gold labels for the data.
    y_pred : numpy.ndarray
        The predicted/observed labels for the data.

    Returns
    -------
    float
        The scored result from the given scorer.
    """
    try:
        scorer = get_scorer(func_name)
    except ValueError:
        scorer = _CUSTOM_METRICS[func_name]

    return scorer._sign * scorer._score_func(y_true, y_pred, **scorer._kwargs)


# a dictionary that maps pre-defined custom metric names to their scorer functions
# this is a private variable only meant for internal use
_PREDEFINED_CUSTOM_METRICS = {
    "f1_score_micro": make_scorer(f1_score, average="micro"),
    "f1_score_macro": make_scorer(f1_score, average="macro"),
    "f1_score_weighted": make_scorer(f1_score, average="weighted"),
    "f1_score_least_frequent": make_scorer(f1_score_least_frequent),
    "f05": make_scorer(fbeta_score, beta=0.5, average="binary"),
    "f05_score_micro": make_scorer(fbeta_score, beta=0.5, average="micro"),
    "f05_score_macro": make_scorer(fbeta_score, beta=0.5, average="macro"),
    "f05_score_weighted": make_scorer(fbeta_score, beta=0.5, average="weighted"),
    "pearson": make_scorer(correlation, corr_type="pearson"),
    "spearman": make_scorer(correlation, corr_type="spearman"),
    "kendall_tau": make_scorer(correlation, corr_type="kendall_tau"),
    "unweighted_kappa": make_scorer(kappa),
    "quadratic_weighted_kappa": make_scorer(kappa, weights="quadratic"),
    "linear_weighted_kappa": make_scorer(kappa, weights="linear"),
    "qwk_off_by_one": make_scorer(kappa, weights="quadratic", allow_off_by_one=True),
    "lwk_off_by_one": make_scorer(kappa, weights="linear", allow_off_by_one=True),
    "uwk_off_by_one": make_scorer(kappa, allow_off_by_one=True),
}

# now create a new dictionary that contains all of the above metrics but
# will also contain any user-defined custom metrics
_CUSTOM_METRICS = copy.deepcopy(_PREDEFINED_CUSTOM_METRICS)



#!/usr/bin/env python
# License: BSD 3 clause
"""
Compute additional evaluation metrics from predictions.

:author: Michael Heilman (mheilman@ets.org)
:author: Nitin Madnani (nmadnani@ets.org)
"""

import argparse
import csv
import logging
from pathlib import Path
from typing import Dict, List, Optional, Union

import numpy as np

from skll.data import Reader, safe_float
from skll.metrics import use_score_func
from skll.types import PathOrStr
from skll.version import __version__

# Make warnings from built-in warnings module get formatted more nicely
logging.captureWarnings(True)
logging.basicConfig(format="%(asctime)s - %(name)s - %(levelname)s - " "%(message)s")
logger = logging.getLogger(__name__)


def get_prediction_from_probabilities(
    classes: Union[List[int], List[str]],
    probabilities: List[float],
    prediction_method: str,
    random_state: int = 1234567890,
) -> Union[int, str]:
    """
    Convert a list of class-probabilities into a class prediction.

    If the prediction method is ``"expected_value"``, the class labels must be integers.

    Parameters
    ----------
    classes: Union[List[int], List[str]]
        List of string or integer class names.
    probabilities: List[float]
        List of probabilities for the respective classes.
    prediction_method: str
        Indicates how to get a single class prediction from the probabilities.
        Must be one of:
            1. ``"highest"``: Selects the class with the highest probability. If
               multiple classes have the same probability, a class is selected randomly.
            2. ``"expected_value"``: Calculates an expected value over integer classes and
               rounds to the nearest int.
    random_state: int, default=1234567890
        Seed for ``numpy.random.RandomState``, used for randomly selecting a class
        when necessary.

    Returns
    -------
    Union[int, str]
        The predicted class.

    Raises
    ------
    ValueError
        If ``classes`` does not contain integers and ``prediction_method`` is
        ``"expected_value"``.

    """
    prng = np.random.RandomState(random_state)
    if prediction_method == "highest":
        highest_p = max(probabilities)
        best_classes = [classes[i] for i, p in enumerate(probabilities) if p == highest_p]
        if len(best_classes) > 1:
            ans = prng.choice(best_classes)
        else:
            ans = best_classes[0]
    elif prediction_method == "expected_value":
        exp_val = 0.0
        for class_, probability in zip(classes, probabilities):
            exp_val += class_ * probability  # type: ignore
        ans = int(round(exp_val))

    return ans


def compute_eval_from_predictions(
    examples_file: PathOrStr,
    predictions_file: PathOrStr,
    metric_names: List[str],
    prediction_method: Optional[str] = None,
) -> Dict[str, float]:
    """
    Compute evaluation metrics from prediction files after running an experiment.

    Parameters
    ----------
    examples_file: :class:`skll.types.PathOrStr`
        Path to a SKLL examples file (in .jsonlines or other format).
    predictions_file: :class:`skll.types.PathOrStr`
        Path to a SKLL predictions output TSV file with id and prediction column names.
    metric_names: List[str]
        A list of SKLL metric names (e.g., ``["pearson", "unweighted_kappa"]``).
    prediction_method: Optional[str], default=None
        Indicates how to get a single class prediction from the probabilities.
        Currently supported options are ``"highest"``, which selects the class
        with the highest probability, and ``"expected_value"``, which calculates
        an expected value over integer classes and rounds to the nearest int.
        If predictions file does not contain probabilities, this should be set
        to ``None``.

    Returns
    -------
    Dict[str, float]
        Mapping of metrics names to corresponding values.

    Raises
    ------
    ValueError
        If the requested prediction method is ``"expected_value"`` but
        the class names can't be converted to ints.
    """
    # convert the examples file and predictions file to a Path
    examples_file = Path(examples_file)
    predictions_file = Path(predictions_file)

    # read gold standard labels
    data = Reader.for_path(examples_file).read()
    if data.labels is not None:
        gold = dict(zip(data.ids, data.labels))

    # read predictions
    pred = {}
    with open(predictions_file) as pred_file:
        reader = csv.reader(pred_file, dialect=csv.excel_tab)
        header = next(reader)

        # If there are more than two columns, assume column 0 contains the ids, and
        # columns 1-n contain class probabilities. Convert them to a class prediction
        # using the specified `method`.
        classes: Union[List[int], List[str]]
        if len(header) > 2:
            classes = [c for c in header[1:] if c]
            if prediction_method is None:
                prediction_method = "highest"
                logger.info("No prediction method specified. Using 'highest'.")
            if prediction_method == "expected_value":
                try:
                    classes = [int(c) for c in classes]
                except ValueError as e:
                    raise e
            for row in reader:
                probabilities = [safe_float(p) for p in row[1:]]
                prediction = get_prediction_from_probabilities(
                    classes, probabilities, prediction_method  # type: ignore
                )
                pred[row[0]] = safe_float(prediction)
        else:
            if prediction_method is not None:
                logger.warning(
                    "A prediction method was provided, but the "
                    "predictions file doesn't contain "
                    "probabilities. Ignoring prediction method "
                    f"'{prediction_method}'."
                )

            for row in reader:
                pred[row[0]] = safe_float(row[1])

    # make a sorted list of example ids in order to match up
    # labels and predictions
    if set(gold.keys()) != set(pred.keys()):
        raise ValueError("The example and prediction IDs do not match.")
    example_ids = sorted(gold.keys())

    res = {}
    for metric_name in metric_names:
        score = use_score_func(
            metric_name,
            np.array([gold[ex_id] for ex_id in example_ids]),
            np.array([pred[ex_id] for ex_id in example_ids]),
        )
        res[metric_name] = score
    return res


def main(argv: Optional[List[str]] = None) -> None:
    """
    Handle command line arguments and get things started.

    Parameters
    ----------
    argv: Optional[List[str]], default=None
        List of arguments, as if specified on the command-line. If ``None``,
        then ``sys.argv[1:]`` is used instead.
    """
    # Get command line arguments
    parser = argparse.ArgumentParser(
        description="Computes evaluation metrics from prediction files after "
        "you have run an experiment.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument("examples_file", help="SKLL input file with labeled examples")
    parser.add_argument("predictions_file", help="file with predictions from SKLL")
    parser.add_argument("metric_names", help="metrics to compute", nargs="+")
    parser.add_argument(
        "--method",
        "-m",
        help="How to generate a prediction from the class "
        "probabilities. Supported methods are 'highest' "
        "(default) and 'expected_value' (only works with"
        " integer classes).",
    )
    parser.add_argument("--version", action="version", version=f"%(prog)s {__version__}")
    args = parser.parse_args(argv)

    supported_prediction_methods = {"highest", "expected_value"}
    if (args.method is not None) and (args.method not in supported_prediction_methods):
        raise KeyError(
            f"Unrecognized prediction method '{args.method}'. "
            "Supported methods are 'highest' and "
            "'expected_value'."
        )

    scores = compute_eval_from_predictions(
        args.examples_file, args.predictions_file, args.metric_names, args.method
    )

    for metric_name in args.metric_names:
        print(f"{scores[metric_name]}\t{metric_name}\t{args.predictions_file}")


if __name__ == "__main__":
    main()



# License: BSD 3 clause
"""
Easily run common scikit-learn experiments with pre-generated features.

:author: Nitin Madnani (nmadnani@ets.org)
:author: Dan Blanchard (dblanchard@ets.org)
:author: Michael Heilman (mheilman@ets.org)
:organization: ETS
"""



#!/usr/bin/env python
# License: BSD 3 clause
"""
Script that converts feature files from one format to another.

:author: Nitin Madnani (nmadnani@ets.org)
:organization: ETS
"""

import argparse
import logging
import sys
from pathlib import Path
from typing import Dict, List, Optional, Tuple

from bs4 import UnicodeDammit

from skll.data.dict_vectorizer import DictVectorizer
from skll.data.readers import EXT_TO_READER
from skll.data.writers import (
    EXT_TO_WRITER,
    ARFFWriter,
    CSVWriter,
    LibSVMWriter,
    TSVWriter,
)
from skll.version import __version__


def _pair_to_dict_tuple(pair: str) -> Tuple[str, int]:
    """Construct mappings from feature/class names to numbers."""
    number_str, name = pair.split("=")
    number = int(number_str)
    return (name, number)


def main(argv: Optional[List[str]] = None) -> None:
    """
    Handle command line arguments and gets things started.

    Parameters
    ----------
    argv : Optional[List[str]], default=None
        List of arguments, as if specified on the command-line.
        If ``None``, ``sys.argv[1:]`` is used instead.
    """
    # Get command line arguments
    parser = argparse.ArgumentParser(
        description="Takes an input feature file and converts it to another "
        "format. Formats are determined automatically from file "
        "extensions.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument(
        "infile",
        help="input feature file (ends in .arff, .csv, " ".jsonlines, .libsvm, .ndj, or .tsv)",
    )
    parser.add_argument(
        "outfile",
        help="output feature file (ends in .arff, .csv, " ".jsonlines, .libsvm, .ndj, or .tsv)",
    )
    parser.add_argument(
        "-i",
        "--id_col",
        help="Name of the column which contains the instance " "IDs in ARFF, CSV, or TSV files.",
        default="id",
    )
    label_group = parser.add_mutually_exclusive_group(required=False)
    label_group.add_argument(
        "-l",
        "--label_col",
        help="Name of the column which contains the "
        "class labels in ARFF, CSV, or TSV files. "
        "For ARFF files, this must be the final "
        "column to count as the label.",
        default="y",
    )
    label_group.add_argument(
        "--no_labels",
        action="store_true",
        default=False,
        help="Used to indicate that the input data has " "no labels.",
    )
    parser.add_argument(
        "-q", "--quiet", help='Suppress printing of "Loading..." messages.', action="store_true"
    )
    parser.add_argument(
        "--arff_regression",
        help="Create ARFF files for regression, not " "classification.",
        action="store_true",
    )
    parser.add_argument(
        "--arff_relation", help="Relation name to use for ARFF file.", default="skll_relation"
    )
    parser.add_argument(
        "--reuse_libsvm_map",
        help="If you want to output multiple files that use "
        "the same mapping from labels and features to "
        "numbers when writing libsvm files, you can "
        "specify an existing .libsvm file to reuse the "
        "mapping from.",
        type=argparse.FileType("rb"),
    )
    parser.add_argument("--version", action="version", version=f"%(prog)s {__version__}")
    args = parser.parse_args(argv)

    # Make warnings from built-in warnings module get formatted more nicely
    logging.captureWarnings(True)
    logging.basicConfig(format="%(asctime)s - %(name)s - %(levelname)s - " "%(message)s")
    logger = logging.getLogger(__name__)

    # make sure the input file extension is one we can process
    input_extension = Path(args.infile).suffix.lower()
    output_extension = Path(args.outfile).suffix.lower()

    if input_extension not in EXT_TO_READER:
        logger.error(
            "Input file must be in either .arff, .csv, .jsonlines, "
            ".libsvm, .ndj, or .tsv format. You specified: "
            f"{input_extension}"
        )
        sys.exit(1)

    # Build feature and label vectorizers from existing libsvm file if asked
    feat_map: Dict[str, int]
    label_map: Optional[Dict[str, int]]
    if args.reuse_libsvm_map and output_extension == ".libsvm":
        feat_map = {}
        label_map = {}
        for line in args.reuse_libsvm_map:
            line = UnicodeDammit(line, ["utf-8", "windows-1252"]).unicode_markup
            if "#" not in line:
                logger.error(
                    "The LibSVM file you want to reuse the map from "
                    "was not created by SKLL and does not actually "
                    "contain the necessary mapping info."
                )
                sys.exit(1)
            comments = line.split("#")[1]
            _, label_map_str, feat_map_str = comments.split("|")
            feat_map.update(_pair_to_dict_tuple(pair) for pair in feat_map_str.strip().split())
            label_map.update(_pair_to_dict_tuple(pair) for pair in label_map_str.strip().split())
        feat_vectorizer = DictVectorizer()
        feat_vectorizer.fit([{name: 1} for name in feat_map])
        feat_vectorizer.vocabulary_ = feat_map
    else:
        feat_vectorizer = None
        label_map = None

    label_col = None if args.no_labels else args.label_col

    # Iterate through input file and collect the information we need
    reader = EXT_TO_READER[input_extension](
        args.infile, quiet=args.quiet, label_col=label_col, id_col=args.id_col
    )
    feature_set = reader.read()
    # write out the file in the requested output format
    writer_type = EXT_TO_WRITER[output_extension]
    writer_args = {"quiet": args.quiet}
    if writer_type is CSVWriter or writer_type is TSVWriter:
        writer_args["label_col"] = label_col
        writer_args["id_col"] = args.id_col
    elif writer_type is ARFFWriter:
        writer_args["label_col"] = label_col
        writer_args["id_col"] = args.id_col
        writer_args["regression"] = args.arff_regression
        writer_args["relation"] = args.arff_relation
    elif writer_type is LibSVMWriter:
        writer_args["label_map"] = label_map
    writer = writer_type(args.outfile, feature_set, **writer_args)
    writer.write()


if __name__ == "__main__":
    main()



# License: BSD 3 clause
"""
Functions related to logging in SKLL.

:author: Nitin Madnani (nmadnani@ets.org)
:organization: ETS
"""
import logging
import re
import warnings
from functools import partial
from os.path import sep
from typing import Optional

orig_showwarning = warnings.showwarning
SKLEARN_WARNINGS_RE = re.compile(re.escape(f"{sep}sklearn{sep}"))


class MatplotlibCategoryFilter(logging.Filter):
    """
    Class to filter out specific log records from `matplotlib.category`.

    This is useful when generating learning curves which generates unnecessary
    log records from `matplotlib.category`. For more details, see this issue:
    https://github.com/matplotlib/matplotlib/issues/23422
    """

    def filter(self, record):
        """
        Implement the filter method.

        Parameters
        ----------
        record : logging.LogRecord
            The log record to be filtered.
        """
        # Check if the log record is from matplotlib.category and contains the specific message
        if (
            record.name == "matplotlib.category"
            and "Using categorical units to plot a list of strings" in record.msg
        ):
            # filter out this record
            return False

        # allow other records through
        return True


def send_sklearn_warnings_to_logger(
    logger, message, category, filename, lineno, file=None, line=None
):
    """
    Return method that sends `sklearn`-specific warnings to a logger.

    This method that can be used to replace warnings.showwarning (via `partial`,
    specifying a `logger` instance).
    """
    if SKLEARN_WARNINGS_RE.search(filename):
        logger.warning(f"{filename}:{lineno}: {category.__name__}:{message}")
    else:
        orig_showwarning(message, category, filename, lineno, file=file, line=line)


def get_skll_logger(
    name: str, filepath: Optional[str] = None, log_level: int = logging.INFO
) -> logging.Logger:
    """
    Create and return logger instances appropriate for use in SKLL code.

    These logger instances can log to both STDERR as well as a file. This
    function will try to reuse any previously created logger based on the
    given name and filepath.

    Parameters
    ----------
    name : str
        The name to be used for the logger.
    filepath : Optional[str], default=None
        The file to be used for the logger via a FileHandler.
        Default: None in which case no file is attached to the
        logger.
    log_level : int, default=logging.INFO
        The level for logging messages

    Returns
    -------
    logger: logging.Logger
        A ``Logger`` instance.
    """
    # first get the logger instance associated with the
    # given name if one already exists
    logger = logging.getLogger(name)
    logger.setLevel(log_level)

    # if we are given a file path and this existing logger doesn't already
    # have a file handler for this file, then add one.
    if filepath:

        def is_file_handler(handler):
            return isinstance(handler, logging.FileHandler) and handler.stream.name == filepath

        need_file_handler = not any([is_file_handler(handler) for handler in logger.handlers])
        if need_file_handler:
            formatter = logging.Formatter("%(asctime)s - %(levelname)s - " "%(message)s")
            file_handler = logging.FileHandler(filepath, mode="w")
            file_handler.setFormatter(formatter)
            file_handler.setLevel(log_level)
            logger.addHandler(file_handler)

    warnings.showwarning = partial(send_sklearn_warnings_to_logger, logger)

    # return the logger instance
    return logger


def close_and_remove_logger_handlers(logger: logging.Logger) -> None:
    """
    Close and remove any handlers attached to a logger instance.

    Parameters
    ----------
    logger : logging.Logger
        Logger instance
    """
    for handler in logger.handlers[:]:
        handler.close()
        logger.removeHandler(handler)



"""
Constants useful for SKLL learners.

:author: Nitin Madnani (nmadnani@ets.org)
:author: Michael Heilman (mheilman@ets.org)
:author: Dan Blanchard (dblanchard@ets.org)
:author: Aoife Cahill (acahill@ets.org)
:organization: ETS
"""

from __future__ import annotations

import os
from typing import Any, Tuple

from sklearn.dummy import DummyClassifier, DummyRegressor
from sklearn.ensemble import (
    AdaBoostClassifier,
    AdaBoostRegressor,
    BaggingClassifier,
    BaggingRegressor,
    GradientBoostingClassifier,
    GradientBoostingRegressor,
    HistGradientBoostingClassifier,
    HistGradientBoostingRegressor,
    RandomForestClassifier,
    RandomForestRegressor,
)
from sklearn.linear_model import (
    BayesianRidge,
    ElasticNet,
    HuberRegressor,
    Lars,
    Lasso,
    LinearRegression,
    LogisticRegression,
    RANSACRegressor,
    Ridge,
    RidgeClassifier,
    SGDClassifier,
    SGDRegressor,
    TheilSenRegressor,
)
from sklearn.naive_bayes import MultinomialNB
from sklearn.neighbors import KNeighborsClassifier, KNeighborsRegressor
from sklearn.neural_network import MLPClassifier, MLPRegressor
from sklearn.svm import SVC, SVR, LinearSVC, LinearSVR
from sklearn.tree import DecisionTreeClassifier, DecisionTreeRegressor

KNOWN_DEFAULT_PARAM_GRIDS = {
    AdaBoostClassifier: {"learning_rate": [0.01, 0.1, 1.0, 10.0, 100.0]},
    AdaBoostRegressor: {"learning_rate": [0.01, 0.1, 1.0, 10.0, 100.0]},
    BaggingClassifier: {
        "max_samples": [0.1, 0.25, 0.5, 1.0],
        "max_features": [0.1, 0.25, 0.5, 1.0],
    },
    BaggingRegressor: {
        "max_samples": [0.1, 0.25, 0.5, 1.0],
        "max_features": [0.1, 0.25, 0.5, 1.0],
    },
    BayesianRidge: {
        "alpha_1": [1e-6, 1e-4, 1e-2, 1, 10],
        "alpha_2": [1e-6, 1e-4, 1e-2, 1, 10],
        "lambda_1": [1e-6, 1e-4, 1e-2, 1, 10],
        "lambda_2": [1e-6, 1e-4, 1e-2, 1, 10],
    },
    DecisionTreeClassifier: {"max_features": ["sqrt", None]},
    DecisionTreeRegressor: {"max_features": ["sqrt", None]},
    DummyClassifier: {},
    DummyRegressor: {},
    ElasticNet: {"alpha": [0.01, 0.1, 1.0, 10.0, 100.0]},
    GradientBoostingClassifier: {"max_depth": [1, 3, 5]},
    GradientBoostingRegressor: {"max_depth": [1, 3, 5]},
    HistGradientBoostingClassifier: {
        "learning_rate": [0.01, 0.1, 1.0],
        "min_samples_leaf": [10, 20, 40],
    },
    HistGradientBoostingRegressor: {
        "loss": ["squared_error", "absolute_error", "poisson"],
        "learning_rate": [0.01, 0.1, 1.0],
        "min_samples_leaf": [10, 20, 40],
    },
    HuberRegressor: {
        "epsilon": [1.05, 1.35, 1.5, 2.0, 2.5, 5.0],
        "alpha": [1e-4, 1e-3, 1e-2, 1e-1, 1, 10, 100, 1000],
    },
    KNeighborsClassifier: {"n_neighbors": [1, 5, 10, 100], "weights": ["uniform", "distance"]},
    KNeighborsRegressor: {"n_neighbors": [1, 5, 10, 100], "weights": ["uniform", "distance"]},
    MLPClassifier: {
        "activation": ["logistic", "tanh", "relu"],
        "alpha": [1e-4, 1e-3, 1e-2, 1e-1, 1],
        "learning_rate_init": [0.001, 0.01, 0.1],
    },
    MLPRegressor: {
        "activation": ["logistic", "tanh", "relu"],
        "alpha": [1e-4, 1e-3, 1e-2, 1e-1, 1],
        "learning_rate_init": [0.001, 0.01, 0.1],
    },
    MultinomialNB: {"alpha": [0.1, 0.25, 0.5, 0.75, 1.0]},
    Lars: {},
    Lasso: {"alpha": [0.01, 0.1, 1.0, 10.0, 100.0]},
    LinearRegression: {},
    LinearSVC: {"C": [0.01, 0.1, 1.0, 10.0, 100.0]},
    LogisticRegression: {"C": [0.01, 0.1, 1.0, 10.0, 100.0]},
    SVC: {
        "C": [0.01, 0.1, 1.0, 10.0, 100.0],
        "gamma": ["auto", "scale", 0.01, 0.1, 1.0, 10.0, 100.0],
    },
    RandomForestClassifier: {"max_depth": [1, 5, 10, None]},
    RandomForestRegressor: {"max_depth": [1, 5, 10, None]},
    RANSACRegressor: {},
    Ridge: {"alpha": [0.01, 0.1, 1.0, 10.0, 100.0]},
    RidgeClassifier: {"alpha": [0.01, 0.1, 1.0, 10.0, 100.0]},
    SGDClassifier: {
        "alpha": [0.000001, 0.00001, 0.0001, 0.001, 0.01],
        "penalty": ["l1", "l2", "elasticnet"],
    },
    SGDRegressor: {
        "alpha": [0.000001, 0.00001, 0.0001, 0.001, 0.01],
        "penalty": ["l1", "l2", "elasticnet"],
    },
    LinearSVR: {"C": [0.01, 0.1, 1.0, 10.0, 100.0]},
    SVR: {
        "C": [0.01, 0.1, 1.0, 10.0, 100.0],
        "gamma": ["auto", "scale", 0.01, 0.1, 1.0, 10.0, 100.0],
    },
    TheilSenRegressor: {},
}

KNOWN_REQUIRES_DENSE: Tuple[Any, ...] = (
    BayesianRidge,
    HistGradientBoostingClassifier,
    HistGradientBoostingRegressor,
    Lars,
    TheilSenRegressor,
)

MAX_CONCURRENT_PROCESSES = int(os.getenv("SKLL_MAX_CONCURRENT_PROCESSES", "5"))

VALID_FEATURE_SCALING_OPTIONS = frozenset(["both", "none", "with_std", "with_mean"])

VALID_SAMPLERS = frozenset(
    ["Nystroem", "RBFSampler", "SkewedChi2Sampler", "AdditiveChi2Sampler", ""]
)

VALID_TASKS = frozenset(["cross_validate", "evaluate", "learning_curve", "predict", "train"])

#: Set of evaluation metrics only used for classification tasks
CLASSIFICATION_ONLY_METRICS = {
    "accuracy",
    "average_precision",
    "balanced_accuracy",
    "f1",
    "f1_score_least_frequent",
    "f1_score_macro",
    "f1_score_micro",
    "f1_score_weighted",
    "f05",
    "f05_score_macro",
    "f05_score_micro",
    "f05_score_weighted",
    "jaccard",
    "jaccard_macro",
    "jaccard_micro",
    "jaccard_weighted",
    "neg_log_loss",
    "precision",
    "precision_macro",
    "precision_micro",
    "precision_weighted",
    "recall",
    "recall_macro",
    "recall_micro",
    "recall_weighted",
    "roc_auc",
}


#: Set of evaluation metrics based on correlation
CORRELATION_METRICS = {"kendall_tau", "pearson", "spearman"}

#: Set of evaluation metrics that can use prediction probabilities
PROBABILISTIC_METRICS = frozenset(["average_precision", "neg_log_loss", "roc_auc"])

#: Set of evaluation metrics only used for regression tasks
REGRESSION_ONLY_METRICS = {
    "explained_variance",
    "max_error",
    "neg_mean_absolute_error",
    "neg_mean_squared_error",
    "neg_root_mean_squared_error",
    "r2",
}

#: Set of unweighted kappa agreement metrics
UNWEIGHTED_KAPPA_METRICS = {"unweighted_kappa", "uwk_off_by_one"}

#: Set of weighed kappa agreement metrics
WEIGHTED_KAPPA_METRICS = {
    "linear_weighted_kappa",
    "lwk_off_by_one",
    "quadratic_weighted_kappa",
    "qwk_off_by_one",
}



# License: BSD 3 clause
"""
Utility classes and functions for running SKLL experiments.

:author: Nitin Madnani (nmadnani@ets.org)
:author: Dan Blanchard (dblanchard@ets.org)
:author: Michael Heilman (mheilman@ets.org)
"""

import json
import math
import re
from collections import defaultdict
from typing import Any, Dict, List, Union

import numpy as np
from sklearn.base import BaseEstimator
from sklearn.pipeline import Pipeline
from tabulate import tabulate  # type: ignore

from skll.types import EvaluateTaskResults, LabelType
from skll.utils.logging import get_skll_logger


class NumpyTypeEncoder(json.JSONEncoder):
    """
    Serialize results to JSON in a numpy-compatible way.

    This class is used when serializing results, particularly the input label
    values if the input has int-valued labels.  Numpy int64 objects can't
    be serialized by the json module, so we must convert them to int objects.

    A related issue where this was adapted from:
    https://stackoverflow.com/questions/11561932/why-does-json-dumpslistnp-arange5-fail-while-json-dumpsnp-arange5-tolis
    """

    def default(self, obj):
        """Encode given object."""
        if isinstance(obj, (np.int32, np.int64)):
            return int(obj)
        elif isinstance(obj, np.ndarray):
            return obj.tolist()
        return json.JSONEncoder.default(self, obj)


class PipelineTypeEncoder(json.JSONEncoder):
    """Serialize ``sklearn.pipeline.Pipeline`` objects."""

    def default(self, obj):
        """Encode given pipeline."""
        if isinstance(obj, Pipeline):
            pipeline_steps = str(obj.named_steps)
            pipeline_steps = re.sub(r"\n", "", pipeline_steps)
            pipeline_steps = re.sub(r"\s+", " ", pipeline_steps)
            return pipeline_steps
        return json.JSONEncoder.default(self, obj)


def _check_job_results(job_results: List[List[Dict[str, Any]]]) -> None:
    """
    See if we have a complete results dictionary for every job.

    Parameters
    ----------
    job_results : List[List[Dict[str, Any]]]
        A list of job result dictionaries.
    """
    logger = get_skll_logger("experiment")
    logger.info("Checking job results")
    for result_dicts in job_results:
        if not result_dicts or "task" not in result_dicts[0]:
            logger.error("There was an error running the experiment:\n" f"{result_dicts}")


def _create_learner_result_dicts(
    task_results: List[EvaluateTaskResults],
    grid_scores: Union[List[None], List[float]],
    grid_search_cv_results_dicts: Union[List[None], List[Dict[str, Any]]],
    learner_result_dict_base: Dict[str, Any],
) -> List[Dict[str, Any]]:
    """
    Create result dictionaries used to write JSON and plain-text results.

    Parameters
    ----------
    task_results : List[:class:`skll.types.EvaluateTaskResults`]
        The task results list.
    grid_scores : Union[List[None], List[float]]
        List of grid scores or a list of ``None`` instances for tasks that do not
        involve grid search.
    grid_search_cv_results_dicts : Union[List[None], List[Dict[str, Any]]]
        A list of dictionaries of grid search CV results, one per fold,
        with keys such as "params", "mean_test_score", etc, that are
        mapped to values associated with each hyperparameter set
        combination. Or a list of ``None`` instances for tasks that do not
        involve any grid search.
    learner_result_dict_base : Dict[str, Any]
        Base dictionary for all learner results.

    Returns
    -------
    res : List[Dict[str, Any]]
        The results of the learners, as a list of dictionaries.
    """
    res = []

    num_folds = len(task_results)
    accuracy_sum = 0.0
    pearson_sum = 0.0
    additional_metric_score_sums: Dict[str, float] = {}
    score_sum = None
    prec_sum_dict: Dict[LabelType, float] = defaultdict(float)
    recall_sum_dict: Dict[LabelType, float] = defaultdict(float)
    f_sum_dict: Dict[LabelType, float] = defaultdict(float)
    result_table = None

    for k, (
        (conf_matrix, fold_accuracy, result_dict, model_params, score, additional_scores),
        grid_score,
        grid_search_cv_results,
    ) in enumerate(zip(task_results, grid_scores, grid_search_cv_results_dicts), start=1):
        # create a new dict for this fold
        learner_result_dict = {}
        learner_result_dict.update(learner_result_dict_base)

        # initialize some variables to blanks so that the
        # set of columns is fixed.
        learner_result_dict["result_table"] = ""
        learner_result_dict["accuracy"] = ""
        learner_result_dict["pearson"] = ""
        learner_result_dict["score"] = ""
        learner_result_dict["fold"] = ""

        if learner_result_dict_base["task"] == "cross_validate":
            learner_result_dict["fold"] = k

        # before we dump the model parameters to disk, we need to handle
        # certain ensemble meta-estimators that have a base estimator
        # since the estimator object is not JSON-serializable; for these
        # meta-estimators, we convert the "estimator" parameter back to
        # a string
        if (estimator := model_params.get("estimator")) and isinstance(estimator, BaseEstimator):
            model_params["estimator"] = estimator.__class__.__name__

        # include model parameters dump for regular learners only;
        # we need to use a special JSON encoder for voting learners
        # that contain ``Pipeline`` objects
        learner_result_dict["model_params"] = json.dumps(model_params, cls=PipelineTypeEncoder)
        if grid_score is not None:
            learner_result_dict["grid_score"] = grid_score
            learner_result_dict["grid_search_cv_results"] = grid_search_cv_results

        if conf_matrix:
            learner_result_dict["conf_matrix"] = conf_matrix
            labels = sorted(task_results[0][2].keys())
            headers = [""] + labels + ["Precision", "Recall", "F-measure"]
            rows = []
            label_metrics = {}
            for i, actual_label in enumerate(labels):
                conf_matrix_row = (
                    conf_matrix[i][:i] + [str([conf_matrix[i][i]])] + conf_matrix[i][i + 1 :]
                )
                label_prf_dict: Dict[str, float] = result_dict[actual_label]
                label_prec = _get_stat_float(label_prf_dict, "Precision")
                label_recall = _get_stat_float(label_prf_dict, "Recall")
                label_f = _get_stat_float(label_prf_dict, "F-measure")
                label_metrics[actual_label] = {
                    "precision": label_prec,
                    "recall": label_recall,
                    "f-measure": label_f,
                }
                if not math.isnan(label_prec):
                    prec_sum_dict[actual_label] += float(label_prec)
                if not math.isnan(label_recall):
                    recall_sum_dict[actual_label] += float(label_recall)
                if not math.isnan(label_f):
                    f_sum_dict[actual_label] += float(label_f)
                result_row = [actual_label] + conf_matrix_row + [label_prec, label_recall, label_f]
                rows.append(result_row)
            learner_result_dict["label_metrics"] = label_metrics
            result_table = tabulate(
                rows, headers=headers, stralign="right", floatfmt=".3f", tablefmt="grid"
            )
            result_table_str = f"{result_table}\n(row = reference; column = " "predicted)"
            learner_result_dict["result_table"] = result_table_str
            if isinstance(fold_accuracy, float):
                learner_result_dict["accuracy"] = fold_accuracy
                accuracy_sum += fold_accuracy

        # if there is no confusion matrix, then we must be dealing
        # with a regression model
        else:
            learner_result_dict.update(result_dict)  # type: ignore
            pearson_sum += float(learner_result_dict["pearson"])

        # get the scores for all the metrics and compute the sums
        if score is not None:
            if score_sum is None:
                score_sum = score
            else:
                score_sum += score
            learner_result_dict["score"] = score
        learner_result_dict["additional_scores"] = additional_scores
        for metric, score in additional_scores.items():
            if score is not None:
                additional_metric_score_sums[metric] = (
                    additional_metric_score_sums.get(metric, 0) + score
                )
        res.append(learner_result_dict)

    if num_folds > 1:
        learner_result_dict = {}
        learner_result_dict.update(learner_result_dict_base)

        learner_result_dict["fold"] = "average"

        if result_table:
            headers = ["Label", "Precision", "Recall", "F-measure"]
            rows = []
            for actual_label in labels:
                # Convert sums to means
                prec_mean = prec_sum_dict[actual_label] / num_folds
                recall_mean = recall_sum_dict[actual_label] / num_folds
                f_mean = f_sum_dict[actual_label] / num_folds
                rows.append([actual_label] + [prec_mean, recall_mean, f_mean])

            result_table = tabulate(rows, headers=headers, floatfmt=".3f", tablefmt="psql")
            learner_result_dict["result_table"] = str(result_table)
            learner_result_dict["accuracy"] = accuracy_sum / num_folds
        else:
            learner_result_dict["pearson"] = pearson_sum / num_folds

        if score_sum is not None:
            learner_result_dict["score"] = score_sum / num_folds
        scoredict = {}
        for metric, score_sum in additional_metric_score_sums.items():
            scoredict[metric] = score_sum / num_folds
        learner_result_dict["additional_scores"] = scoredict
        res.append(learner_result_dict)

    return res


def _get_stat_float(label_result_dict: Dict[str, float], stat: str) -> float:
    """
    Extract precision, recall, and f-score values from the given dictionary.

    Parameters
    ----------
    label_result_dict : Dict[str, float]
        Dictionary containing the stats to retrieve for a particular label.
    stat : str
        The statistic we're looking for in the dictionary.

    Returns
    -------
    float
        The value of the stat if it's in the dictionary, and NaN otherwise.
    """
    if stat in label_result_dict and label_result_dict[stat] is not None:
        return label_result_dict[stat]
    else:
        return float("nan")



"""Utility functions to make SKLL testing simpler."""


import os
import re
from collections import OrderedDict
from math import floor, log10
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple

import numpy as np
from numpy.random import RandomState
from sklearn.datasets import (
    fetch_california_housing,
    load_digits,
    make_classification,
    make_regression,
)
from sklearn.feature_extraction import FeatureHasher
from sklearn.model_selection import KFold, ShuffleSplit, StratifiedKFold
from sklearn.utils import shuffle as sk_shuffle

from skll.config import _setup_config_parser, fix_json
from skll.data import FeatureSet, NDJWriter
from skll.types import PathOrStr

if env_test_dir := os.getenv("TESTDIR"):
    tests_dir = Path(env_test_dir) / "tests"
else:
    tests_dir = Path(__file__).resolve().parent.parent.parent / "tests"
config_dir = tests_dir / "configs"
backward_compatibility_dir = tests_dir / "backward_compatibility"
examples_dir = tests_dir.parent / "examples"
output_dir = tests_dir / "output"
other_dir = tests_dir / "other"
train_dir = tests_dir / "train"
test_dir = tests_dir / "test"


class BoolDict(dict):
    """Dictionary that returns ``False`` instead of ``None`` as default."""

    def __getitem__(self, key):
        """Get item for given key."""
        return super().get(key, False)


def unlink(file_path: PathOrStr):
    """
    Remove a file path if it exists.

    Parameters
    ----------
    file_path : :class:`skll.types.PathOrStr`
        File path to remove.
    """
    file_path = Path(file_path)
    if file_path.exists():
        file_path.unlink()


def fill_in_config_paths(config_template_path: PathOrStr) -> Path:
    """
    Fill paths in given configuration file template.

    Add various paths in the given configuration template file.

    Parameters
    ----------
    config_template_path : :class:`skll.types.PathOrStr`
        The path to the template configuration file.

    Returns
    -------
    Path
        The path to the filled configuration file.
    """
    # convert path to Path object
    config_template_path = Path(config_template_path)

    config = _setup_config_parser(config_template_path, validate=False)

    task = config.get("General", "task")

    config.set("Input", "train_directory", str(train_dir))

    to_fill_in = ["logs"]

    if task != "learning_curve":
        to_fill_in.append("predictions")

    if task not in ["cross_validate", "learning_curve"]:
        to_fill_in.append("models")

    if task in ["cross_validate", "evaluate", "learning_curve", "train"]:
        to_fill_in.append("results")

    for d in to_fill_in:
        config.set("Output", d, str(output_dir))

    if task == "cross_validate":
        folds_file = config.get("Input", "folds_file")
        if folds_file:
            config.set("Input", "folds_file", str(train_dir / folds_file))

    if task == "predict" or task == "evaluate":
        config.set("Input", "test_directory", str(test_dir))

    # set up custom learner path, if relevant
    custom_learner_path = config.get("Input", "custom_learner_path")
    custom_learner_abs_path = tests_dir / custom_learner_path
    config.set("Input", "custom_learner_path", str(custom_learner_abs_path))

    config_prefix_match = re.search(r"^(.*)\.template\.cfg", str(config_template_path))
    assert config_prefix_match is not None
    config_prefix = config_prefix_match.groups()[0]
    new_config_path = Path(f"{config_prefix}.cfg")

    with open(new_config_path, "w") as new_config_file:
        config.write(new_config_file)

    return new_config_path


def fill_in_config_paths_for_single_file(
    config_template_path: PathOrStr,
    train_file: str,
    test_file: str,
    train_directory: str = "",
    test_directory: str = "",
) -> Path:
    """
    Fill in input file and directory paths in given configuration template.

    Add paths to train files, test files, and output directories in the
    given config template file.

    Parameters
    ----------
    config_template_path : :class:`skll.types.PathOrStr`
        Path to the template configuration file.
    train_file : str
        Name of the training data file.
    test_file : str
        Name of the test data file.
    train_directory : str, optional
        Path to the directory containing the training data.
        Defaults to ''
    test_directory : str, optional
        Path to the directory containing the test data.
        Defaults to ''

    Returns
    -------
    Path
        The path to the filled configuration file.
    """
    # convert path to Path object if it's a string
    config_template_path = Path(config_template_path)

    config = _setup_config_parser(config_template_path, validate=False)

    task = config.get("General", "task")

    config.set("Input", "train_file", str(train_dir / train_file))
    if task == "predict" or task == "evaluate":
        config.set("Input", "test_file", str(test_dir / test_file))

    if train_directory:
        config.set("Input", "train_directory", str(train_dir / train_directory))

    if test_directory:
        config.set("Input", "test_directory", str(test_dir / test_directory))

    to_fill_in = ["logs"]
    if task != "train":
        to_fill_in.append("predictions")

    if task == "cross_validate":
        if config.get("Output", "save_cv_models"):
            to_fill_in.append("models")
    else:
        to_fill_in.append("models")

    if task in ["cross_validate", "evaluate", "train"]:
        to_fill_in.append("results")

    for d in to_fill_in:
        config.set("Output", d, str(output_dir))

    if task == "cross_validate":
        folds_file = config.get("Input", "folds_file")
        if folds_file:
            config.set("Input", "folds_file", str(train_dir / folds_file))

    config_prefix_match = re.search(r"^(.*)\.template\.cfg", str(config_template_path))
    assert config_prefix_match is not None
    config_prefix = config_prefix_match.groups()[0]
    new_config_path = Path(f"{config_prefix}.cfg")

    with open(new_config_path, "w") as new_config_file:
        config.write(new_config_file)

    return new_config_path


def fill_in_config_options(
    config_template_path: PathOrStr,
    values_to_fill_dict: Dict[str, Any],
    sub_prefix: str,
    good_probability_option: bool = False,
) -> Path:
    """
    Fill in configuration options in the given template file.

    Parameters
    ----------
    config_template_path : :class:`skll.types.PathOrStr`
        Path to the template configuration file.
    values_to_fill_dict : Dict[str, Any]
        Dictionary containing the options to fille the keys and the
        corresponding values.
    sub_prefix : str
        The sub-prefix to add to the name when creating
        the filled configuration file on disk.
    good_probability_option : bool, default=False
        Whether to add the "probability" option in the correct
        section or an incorrect section.

    Returns
    -------
    Path
        The path to the filled configuration file.
    """
    # convert path to Path object if it's a string
    config_template_path = Path(config_template_path)

    config = _setup_config_parser(config_template_path, validate=False)

    # The dictionary that says which option to fill in where
    # Note: (a) `bad_option` and `duplicate_option` are needed
    # for test_config_parsing_invalid_option and
    # test_config_parsing_duplicate_option() respectively.
    # (b) `probability` is deliberately specified in the wrong
    # section for test_config_parsing_option_in_wrong_section().
    to_fill_in = {
        "General": ["experiment_name", "task"],
        "Input": [
            "train_directory",
            "train_file",
            "test_directory",
            "test_file",
            "featuresets",
            "featureset_names",
            "custom_learner_path",
            "feature_hasher",
            "hasher_features",
            "learners",
            "sampler",
            "shuffle",
            "feature_scaling",
            "learning_curve_cv_folds_list",
            "folds_file",
            "learning_curve_train_sizes",
            "fixed_parameters",
            "num_cv_folds",
            "cv_seed",
            "bad_option",
            "duplicate_option",
            "suffix",
        ],
        "Tuning": [
            "grid_search",
            "objective",
            "min_feature_count",
            "use_folds_file_for_grid_search",
            "grid_search_folds",
            "pos_label",
            "param_grids",
            "objectives",
            "duplicate_option",
        ],
        "Output": [
            "results",
            "log",
            "logs",
            "models",
            "metrics",
            "predictions",
            "pipeline",
            "save_cv_folds",
            "save_cv_models",
            "save_votes",
            "wandb_credentials",
        ],
    }

    if good_probability_option:
        to_fill_in["Output"].append("probability")
    else:
        to_fill_in["Tuning"].append("probability")

    for section in to_fill_in:
        for param_name in to_fill_in[section]:
            if param_name in values_to_fill_dict:
                value = str(values_to_fill_dict[param_name])
                config.set(section, param_name, value)

    config_prefix_match = re.search(r"^(.*)\.template\.cfg", str(config_template_path))
    assert config_prefix_match is not None
    config_prefix = config_prefix_match.groups()[0]
    new_config_path = Path(f"{config_prefix}_{sub_prefix}.cfg")

    with open(new_config_path, "w") as new_config_file:
        config.write(new_config_file)

    return new_config_path


def fill_in_config_paths_for_fancy_output(config_template_path: PathOrStr) -> Path:
    """
    Fill in the template for more comprehensive ("fancier") output.

    Add paths to train, test, and output directories in the given
    config template file.

    Parameters
    ----------
    config_template_path : :class:`skll.types.PathOrStr`
        Path to the template configuration file.

    Returns
    -------
    Path
        The path to the filled configuration file.
    """
    # convert template path to a Path object if string
    config_template_path = Path(config_template_path)

    config = _setup_config_parser(config_template_path, validate=False)

    config.set("Input", "train_file", str(train_dir / "fancy_train.jsonlines"))
    config.set("Input", "test_file", str(test_dir / "fancy_test.jsonlines"))
    config.set("Output", "results", str(output_dir))
    config.set("Output", "logs", str(output_dir))
    config.set("Output", "predictions", str(output_dir))

    config_prefix_match = re.search(r"^(.*)\.template\.cfg", str(config_template_path))
    assert config_prefix_match is not None
    config_prefix = config_prefix_match.groups()[0]
    new_config_path = Path(f"{config_prefix}.cfg")

    with open(new_config_path, "w") as new_config_file:
        config.write(new_config_file)

    return new_config_path


def fill_in_config_options_for_voting_learners(
    learner_type: str, task: str, options_dict: BoolDict
) -> Tuple:
    """
    Fill in values specific to voting learners in the given template.

    Parameters
    ----------
    learner_type : str
        One of "classifier" or "regressor".
    task : str
        The voting learner task for which we are filling the values.
    options_dict : BoolDict
        A dictionary containing various boolean options which will
        be used to fill the values based on whether they are set
        to ``True`` or ``False`.

    Returns
    -------
    res : a 13-tuple
        A tuple containing the following items as determined by the options
        specified in ``options_dict``:
        - the path to the filled-in configuration file
        - the names of the underlying estimators used for the voting learner
        - a job prefix constructed using the experiment name, the task, the
          featureset name, and the learner name.
        - the path to any custom learner file used, if any
        - the names of the objective functions used, if any
        - the names of any output metrics used, if any
        - the list of fixed parameters used for the underlying estimators, if any
        - the list of parameter grids used for the underlying estimators, if any
        - the list of samplers used for the underlying estimators, if any
        - the number of cross-validation folds used (6 or 10)
        - the custom seed value used for cross-validation, if any
        - the number of learning curve cross-validation folds (10 or 20)
        - the list of learning curve training sizes
    """
    # setup learner-type specific values based on configuration options
    custom_learner = ""
    objectives = None
    output_metrics = []
    model_kwargs_list: Optional[List[Dict[str, Any]]] = None
    param_grid_list: Optional[List[Dict[str, Any]]] = None
    sampler_list = None
    num_cv_folds = learning_curve_cv_folds = 10
    cv_seed = 123456789
    fixed_parameters: List[Dict[str, Any]]
    learning_curve_train_sizes = [0.1, 0.325, 0.55, 0.775, 1.0]
    if learner_type == "classifier":
        learner_name = "VotingClassifier"
        estimator_names = ["LogisticRegression", "SVC", "MultinomialNB"]
        fixed_parameters = [{"estimator_names": estimator_names}]

        if options_dict["with_model_kwargs_list"]:
            model_kwargs_list = [{}, {"gamma": "auto"}, {"alpha": 0.01}]
            fixed_parameters[0].update({"estimator_fixed_parameters": model_kwargs_list})

        if options_dict["with_param_grid_list"]:
            param_grid_list = [{"penalty": ["l1", "l2"]}, {}, {}]
            fixed_parameters[0].update({"estimator_param_grids": param_grid_list})

        if options_dict["with_sampler_list"]:
            sampler_list = ["SkewedChi2Sampler", "Nystroem", ""]
            fixed_parameters[0].update({"estimator_samplers": sampler_list})

        if options_dict["with_soft_voting"]:
            fixed_parameters[0].update({"voting_type": "soft"})

        if options_dict["with_grid_search"]:
            objectives = (
                ["accuracy", "f1_score_macro"]
                if options_dict["with_multiple_objectives"]
                else ["accuracy"]
            )

        if options_dict["with_output_metrics"]:
            output_metrics = ["f05", "f1_score_macro"]

    else:
        learner_name = "VotingRegressor"
        estimator_names = ["LinearRegression", "SVR", "RandomForestRegressor"]
        fixed_parameters = [{"estimator_names": estimator_names}]

        if options_dict["with_model_kwargs_list"]:
            model_kwargs_list = [{}, {"kernel": "poly"}, {"n_estimators": 200}]
            fixed_parameters[0].update({"estimator_fixed_parameters": model_kwargs_list})

        if options_dict["with_param_grid_list"]:
            param_grid_list = [{}, {"degree": [3, 4, 5]}, {}]
            fixed_parameters[0].update({"estimator_param_grids": param_grid_list})

        if options_dict["with_sampler_list"]:
            sampler_list = ["", "AdditiveChi2Sampler", ""]
            fixed_parameters[0].update({"estimator_samplers": sampler_list})

        if options_dict["with_grid_search"]:
            objectives = (
                ["pearson", "neg_mean_squared_error"]
                if options_dict["with_multiple_objectives"]
                else ["pearson"]
            )

        if options_dict["with_output_metrics"]:
            output_metrics = ["spearman", "kendall_tau"]

    # create the dictionary that we will use to fill our config template
    experiment_name = "test_voting_learner"
    sub_prefix = task
    featureset_name = "f0"
    train_path = train_dir / f"{featureset_name}.jsonlines"
    job_name = f"{experiment_name}_{sub_prefix}_{learner_name}"
    values_to_fill_dict = {
        "experiment_name": f"{experiment_name}_{sub_prefix}",
        "train_file": str(train_path),
        "task": task,
        "learners": str([learner_name]),
        "grid_search": fix_json(str(options_dict["with_grid_search"])),
        "featureset_names": str([featureset_name]),
        "fixed_parameters": str(fixed_parameters),
    }

    # insert additional values in this dictionary
    if task in ["evaluate", "predict"]:
        test_path = str(test_dir / f"{featureset_name}.jsonlines")
        values_to_fill_dict["test_file"] = str(test_path)

    if options_dict["with_grid_search"]:
        values_to_fill_dict["objectives"] = str(objectives)

    if options_dict["with_gs_folds"]:
        values_to_fill_dict["grid_search_folds"] = "4"

    if options_dict["with_shuffle"]:
        values_to_fill_dict["shuffle"] = "true"

    if options_dict["with_centering"]:
        values_to_fill_dict["feature_scaling"] = "with_mean"

    if options_dict["with_min_feature_count"]:
        values_to_fill_dict["min_feature_count"] = "2"

    if options_dict["with_custom_learner_path"]:
        custom_learner = str(other_dir / "custom_logistic_wrapper.py")
        values_to_fill_dict["custom_learner_path"] = custom_learner

    if options_dict["with_pos_label"]:
        values_to_fill_dict["pos_label"] = "dog"

    if options_dict["with_individual_predictions"]:
        values_to_fill_dict["save_votes"] = "true"

    if options_dict["with_output_metrics"]:
        values_to_fill_dict["metrics"] = str(output_metrics)

    if options_dict["with_prediction_prefix"]:
        values_to_fill_dict["predictions"] = str(output_dir)

    if options_dict["with_cv_folds"]:
        num_cv_folds = 6
        values_to_fill_dict["num_cv_folds"] = str(num_cv_folds)

    if options_dict["with_custom_cv_seed"]:
        cv_seed = 987
        values_to_fill_dict["cv_seed"] = str(cv_seed)

    if options_dict["without_save_cv_folds"]:
        values_to_fill_dict["save_cv_folds"] = "false"

    if options_dict["with_save_cv_models"]:
        values_to_fill_dict["save_cv_models"] = "true"

    if options_dict["with_learning_curve_cv_folds"]:
        learning_curve_cv_folds = 20
        values_to_fill_dict["learning_curve_cv_folds_list"] = str([learning_curve_cv_folds])

    if options_dict["with_learning_curve_train_sizes"]:
        learning_curve_train_sizes = [0.1, 0.5, 1.0]
        values_to_fill_dict["learning_curve_train_sizes"] = str(learning_curve_cv_folds)

    if task != "learning_curve":
        values_to_fill_dict["models"] = (
            str(other_dir) if options_dict["with_existing_model"] else str(output_dir)
        )

    # locate the voting learner config template and instantiate it
    config_template_path = config_dir / f"{experiment_name}.template.cfg"
    config_path = fill_in_config_options(config_template_path, values_to_fill_dict, sub_prefix)

    return (
        config_path,
        estimator_names,
        job_name,
        custom_learner,
        objectives,
        output_metrics,
        model_kwargs_list,
        param_grid_list,
        sampler_list,
        num_cv_folds,
        cv_seed,
        learning_curve_cv_folds,
        learning_curve_train_sizes,
    )


def create_jsonlines_feature_files(path: PathOrStr):
    """
    Create dummy jsonlines feature files and save them under ``path``.

    Parameters
    ----------
    path : :class:`skll.types.PathOrStr`
        Full path under which to save the created feature files.
    """
    # convert to Path object
    path = Path(path)

    # we only need to create the feature files if they
    # don't already exist under the given path
    feature_files_to_create = [path / f"f{i}.jsonlines" for i in range(6)]
    if all([ff.exists() for ff in feature_files_to_create]):
        return
    else:
        num_examples = 1000
        np.random.seed(1234567890)

        # Create lists we will write files from
        ids = []
        features = []
        labels = []
        for j in range(num_examples):
            y = "dog" if j % 2 == 0 else "cat"
            ex_id = f"{y}{j}"
            x = {f"f{feat_num}": np.random.randint(0, 4) for feat_num in range(5)}
            x = OrderedDict(sorted(x.items(), key=lambda t: t[0]))
            ids.append(ex_id)
            labels.append(y)
            features.append(x)

        for i in range(5):
            file_path = path / f"f{i}.jsonlines"
            sub_features = []
            for example_num in range(num_examples):
                feat_num = i
                x = {f"f{feat_num}": features[example_num][f"f{feat_num}"]}
                sub_features.append(x)
            fs = FeatureSet("ablation_cv", ids, features=sub_features, labels=labels)

            writer = NDJWriter(file_path, fs)
            writer.write()

        # now write out the last file which is basically
        # identical to the last featureset we wrote
        # except that it has two extra instances
        fs = FeatureSet(
            "extra",
            ids + [f"cat{num_examples}", f"dog{num_examples + 1}"],
            features=sub_features + [{}, {}],
            labels=labels + ["cat", "dog"],
        )
        file_path = path / "f5.jsonlines"
        writer = NDJWriter(file_path, fs)
        writer.write()


def remove_jsonlines_feature_files(path: PathOrStr):
    """
    Remove all files created by ``create_jsonlines_feature_files()``.

    Parameters
    ----------
    path : :class:`skll.types.PathOrStr`
        Path to directory in which jsonlines files reside.
    """
    for i in range(6):
        unlink(Path(path) / f"f{i}.jsonlines")


def make_classification_data(
    num_examples=100,
    train_test_ratio=0.5,
    num_features=10,
    use_feature_hashing=False,
    feature_bins=4,
    num_labels=2,
    empty_labels=False,
    string_label_list=None,
    feature_prefix="f",
    id_type="string",
    class_weights=None,
    non_negative=False,
    one_string_feature=False,
    num_string_values=4,
    random_state=1234567890,
):
    """
    Create dummy classification data for use in various tests.

    Parameters
    ----------
    num_examples : int, optional
        Number of examples in the generated data.
        Defaults to 100.
    train_test_ratio : float, optional
        Ratio of train to test data in the generated data.
        Defaults to 0.5.
    num_features : int, optional
        Number of features in each generated example.
        Defaults to 10.
    use_feature_hashing : bool, optional
        Whether to use feature hashing.
        Defaults to ``False``.
    feature_bins : int, optional
        How many hashed feature bins to use, if ``use_feature_hashing``
        is ``True``.
        Defaults to 4.
    num_labels : int, optional
        Number of classes to use.
    empty_labels : bool, optional
        Whether to set the training and test labels to ``None``
        for certain tests.
        Defaults to ``False``.
    string_label_list : List[str], optional
        List of pre-specified string labels.
    feature_prefix : str, optional
        The prefix to use for all the feature names, if any.
        Defaults to "f" (i.e., "f01" et cetera).
    id_type : str, optional
        The data types to use for the generated example IDs.
        One of ["float", "integer", "integer_string", "string"].
        Defaults to "string".
    class_weights : List[float], optional
        The proportion of generated samples for each of the classes.
        If ``None``, use the same number of samples for each class.
        Defaults to ``None``.
    non_negative : bool, optional
        Whether to generate only non-negative features.
        Defaults to ``False``.
    one_string_feature : bool, optional
        Whether to generate one feature with string values in
        the generated data.
        Defaults to ``False``.
    num_string_values : int, optional
        If ``one_string_feature`` is set to ``True``, how many
        possible values to use for that feature.
        Defaults to 4.
    random_state : int, optional
        Random state to use for generating the data.
        Defaults to 123456789.

    Returns
    -------
    Tuple
        Tuple containing the generated training featureset and
        the generated test featureset.

    """
    # use sklearn's make_classification to generate the data for us
    num_numeric_features = num_features - 1 if one_string_feature else num_features
    X, y = make_classification(
        n_samples=num_examples,
        n_features=num_numeric_features,
        n_informative=num_numeric_features,
        n_redundant=0,
        n_classes=num_labels,
        weights=class_weights,
        random_state=random_state,
    )

    if string_label_list:
        assert len(string_label_list) == num_labels
        label_to_string = np.vectorize(lambda n: string_label_list[n])
        y = label_to_string(y)

    # if we were told to only generate non-negative features, then
    # we can simply take the absolute values of the generated features
    if non_negative:
        X = abs(X)

    # since we want to use SKLL's FeatureSet class, we need to
    # create a list of IDs; we create IDs that either can also
    # be numbers or pure strings
    if id_type == "string":
        ids = [f"EXAMPLE_{n}" for n in range(1, num_examples + 1)]
    elif id_type == "integer_string":
        ids = [f"{n}" for n in range(1, num_examples + 1)]
    elif id_type == "float":
        ids = [float(n) for n in range(1, num_examples + 1)]
    elif id_type == "integer":
        ids = list(range(1, num_examples + 1))

    # create a string feature that has four possible values
    # 'a', 'b', 'c' and 'd' and add it to X at the end
    if one_string_feature:
        prng = RandomState(random_state)
        random_indices = prng.random_integers(0, num_string_values - 1, num_examples)
        possible_values = [chr(x) for x in range(97, 97 + num_string_values)]
        string_feature_values = [possible_values[i] for i in random_indices]
        string_feature_column = np.array(string_feature_values, dtype=object).reshape(100, 1)
        X = np.append(X, string_feature_column, 1)

    # create a list of dictionaries as the features
    feature_names = [f"{feature_prefix}{n:02d}" for n in range(1, num_features + 1)]
    features = [dict(zip(feature_names, row)) for row in X]

    # split everything into training and testing portions
    num_train_examples = int(round(train_test_ratio * num_examples))
    train_features, test_features = (features[:num_train_examples], features[num_train_examples:])
    train_y, test_y = y[:num_train_examples], y[num_train_examples:]
    train_ids, test_ids = ids[:num_train_examples], ids[num_train_examples:]

    # are we told to generate empty labels
    train_labels = None if empty_labels else train_y
    test_labels = None if empty_labels else test_y

    # create a FeatureHasher if we are asked to use feature hashing
    # with the specified number of feature bins
    vectorizer = FeatureHasher(n_features=feature_bins) if use_feature_hashing else None
    train_fs = FeatureSet(
        "classification_train",
        train_ids,
        labels=train_labels,
        features=train_features,
        vectorizer=vectorizer,
    )
    if train_test_ratio < 1.0:
        test_fs = FeatureSet(
            "classification_test",
            test_ids,
            labels=test_labels,
            features=test_features,
            vectorizer=vectorizer,
        )
    else:
        test_fs = None

    return (train_fs, test_fs)


def make_regression_data(
    num_examples=100,
    train_test_ratio=0.5,
    num_features=2,
    sd_noise=1.0,
    use_feature_hashing=False,
    feature_bins=4,
    start_feature_num=1,
    random_state=1234567890,
):
    """
    Create dummy regression data for use with tests.

    Parameters
    ----------
    num_examples : int, optional
        Number of examples in the generated data.
        Defaults to 100.
    train_test_ratio : float, optional
        Ratio of train to test data in the generated data.
        Defaults to 0.5.
    num_features : int, optional
        Number of features in each generated example.
        Defaults to 2.
    sd_noise : float, optional
        Amount of Gaussian noise added to the output.
        Defaults to 1.0.
    use_feature_hashing : bool, optional
        Whether to use feature hashing.
        Defaults to ``False``.
    feature_bins : int, optional
        How many hashed feature bins to use, if ``use_feature_hashing``
        is ``True``.
        Defaults to 4.
    start_feature_num : int, optional
        The integer suffix for the first feature name (i.e., "f01").
        Defaults to 1.
    random_state : int, optional
        Random state to use for generating the data.
        Defaults to 123456789.

    Returns
    -------
    Tuple
        3-tuple containing the generated training featureset, the generated
        test featureset, and a dictionary containing the oracle feature
        weights
    """
    # if we are doing feature hashing and we have asked for more
    # feature bins than number of total features, we need to
    # handle that because `make_regression()` doesn't know
    # about hashing
    if use_feature_hashing and num_features < feature_bins:
        num_features = feature_bins

    # use sklearn's make_regression to generate the data for us
    X, y, weights = make_regression(
        n_samples=num_examples,
        n_features=num_features,
        noise=sd_noise,
        random_state=random_state,
        coef=True,
    )

    # since we want to use SKLL's FeatureSet class, we need to
    # create a list of IDs
    ids = [f"EXAMPLE_{n}" for n in range(1, num_examples + 1)]

    # create a list of dictionaries as the features
    index_width_for_feature_name = int(floor(log10(num_features))) + 1
    feature_names = []
    for n in range(start_feature_num, start_feature_num + num_features):
        index_str = str(n).zfill(index_width_for_feature_name)
        feature_name = f"f{index_str}"
        feature_names.append(feature_name)
    features = [dict(zip(feature_names, row)) for row in X]

    # At this point the labels are generated using unhashed features
    # even if we want to do feature hashing. `make_regression()` from
    # sklearn doesn't know anything about feature hashing, so we need
    # a hack here to compute the updated labels ourselves
    # using the same command that sklearn uses inside `make_regression()`
    # which is to generate the X and the weights and then compute the
    # y as the dot product of the two. This y will then be used as our
    # labels instead of the original y we got from `make_regression()`.
    # Note that we only want to use the number of weights that are
    # equal to the number of feature bins for the hashing
    if use_feature_hashing:
        feature_hasher = FeatureHasher(n_features=feature_bins)
        hashed_X = feature_hasher.fit_transform(features)
        y = hashed_X.dot(weights[:feature_bins])

    # convert the weights array into a dictionary for convenience
    # if we are using feature hashing, we need to use the names
    # that would be output by `model_params()` instead of the
    # original names since that's what we would get from SKLL
    if use_feature_hashing:
        index_width_for_feature_name = int(floor(log10(feature_bins))) + 1
        hashed_feature_names = []
        for i in range(feature_bins):
            index_str = str(i + 1).zfill(index_width_for_feature_name)
            feature_name = f"hashed_feature_{index_str}"
            hashed_feature_names.append(feature_name)
        weightdict = dict(zip(hashed_feature_names, weights[:feature_bins]))
    else:
        weightdict = dict(zip(feature_names, weights))

    # split everything into training and testing portions
    num_train_examples = int(round(train_test_ratio * num_examples))
    train_features, test_features = (features[:num_train_examples], features[num_train_examples:])
    train_y, test_y = y[:num_train_examples], y[num_train_examples:]
    train_ids, test_ids = ids[:num_train_examples], ids[num_train_examples:]

    # create a FeatureHasher if we are asked to use feature hashing
    # with the specified number of feature bins
    vectorizer = FeatureHasher(n_features=feature_bins) if use_feature_hashing else None
    train_fs = FeatureSet(
        "regression_train",
        train_ids,
        labels=train_y,
        features=train_features,
        vectorizer=vectorizer,
    )
    test_fs = FeatureSet(
        "regression_test", test_ids, labels=test_y, features=test_features, vectorizer=vectorizer
    )

    return (train_fs, test_fs, weightdict)


def make_sparse_data(use_feature_hashing=False):
    """
    Create sparse data for use in various tests.

    Function to create sparse data with two features always zero
    in the training set and a different one always zero in the
    test set.

    Parameters
    ----------
    use_feature_hashing : bool, optional
        Whether to use feature hashing.
        Defaults to ``False``.

    Returns
    -------
    Tuple
        Tuple containing the generated training featureset and
        the generated test featureset.
    """
    # Create training data
    X, y = make_classification(
        n_samples=500,
        n_features=3,
        n_informative=3,
        n_redundant=0,
        n_classes=2,
        random_state=1234567890,
    )

    # we need features to be non-negative since we will be
    # using naive bayes laster
    X = np.abs(X)

    # make sure that none of the features are zero
    X[np.where(X == 0)] += 1

    # since we want to use SKLL's FeatureSet class, we need to
    # create a list of IDs
    ids = [f"EXAMPLE_{n}" for n in range(1, 501)]

    # create a list of dictionaries as the features
    # with f1 and f5 always 0
    feature_names = [f"f{n}" for n in range(1, 6)]
    features = []
    for row in X:
        row = [0] + row.tolist() + [0]
        features.append(dict(zip(feature_names, row)))

    # use a FeatureHasher if we are asked to do feature hashing
    vectorizer = FeatureHasher(n_features=4) if use_feature_hashing else None
    train_fs = FeatureSet("train_sparse", ids, features=features, labels=y, vectorizer=vectorizer)

    # now create the test set with f4 always 0 but nothing else
    X, y = make_classification(
        n_samples=100,
        n_features=4,
        n_informative=4,
        n_redundant=0,
        n_classes=2,
        random_state=1234567890,
    )
    X = np.abs(X)
    X[np.where(X == 0)] += 1
    ids = [f"EXAMPLE_{n}" for n in range(1, 101)]

    # create a list of dictionaries as the features
    # with f4 always 0
    feature_names = [f"f{n}" for n in range(1, 6)]
    features = []
    for row in X:
        row = row.tolist()
        row = row[:3] + [0] + row[3:]
        features.append(dict(zip(feature_names, row)))

    test_fs = FeatureSet("test_sparse", ids, features=features, labels=y, vectorizer=vectorizer)

    return train_fs, test_fs


def make_digits_data(num_examples=None, test_size=0.2, use_digit_names=False):
    """
    Create train/test featuresets from the digits dataset.

    Parameters
    ----------
    num_examples : int, optional
        Number of total examples to use. 80% of these examples
        will be in the training set and the remaining 20% in
        the test set. If ``None``, it will use all of the
        examples in the dataset.
        Defaults to ``None``.
    test_size : float, optional
        Fraction of ``num_examples`` to use for the test
        featureset. Should be between 0 and 1. If this is 0,
        the training featureset will contain all of the
        examples and the test featureset will be ``None``.
        Defaults to 0.2.
    use_digit_names : bool, optional
        If ``True``, use the names of the digits ("zero", "one", etc.) as the
        labels in the featuresets rather than the integer-valued targets
        (0, 1, etc.).
        Defaults to ``False``.

    Returns
    -------
    fs_tuple : tuple
        A 2-tuple containing the created train and test featuresets.
        The test featureset will be ``None`` if ``test_size`` is 0.

    Raises
    ------
    ValueError
        If ``num_examples`` is greater than the number of available
        examples.
    """
    # load the digits data
    digits = load_digits(as_frame=True)
    df_digits = digits.frame

    # use all examples if ``num_examples`` was ``None``
    num_examples = len(df_digits) if num_examples is None else num_examples

    # raise an exception if the number of desired examples is greater
    # than the number of available examples in the data
    if num_examples > len(df_digits):
        raise ValueError(
            f"invalid value {num_examples} for 'num_examples', "
            f"only {len(df_digits)} examples are available."
        )

    # select the desired number of examples
    row_indices = np.arange(len(df_digits))
    prng = np.random.default_rng(123456789)
    if num_examples < len(df_digits):
        chosen_row_indices = prng.choice(
            row_indices, size=num_examples, replace=False, shuffle=False
        )
    else:
        chosen_row_indices = row_indices

    # now split the chosen indices into train and test indices
    # assuming ``test_size`` > 0
    if test_size > 0:
        splitter = ShuffleSplit(n_splits=1, test_size=test_size, random_state=123456789)
        train_indices, test_indices = list(splitter.split(chosen_row_indices))[0]
    else:
        train_indices = chosen_row_indices
        test_indices = np.array([])

    # by default, we will use the "target" column of the frame as our labels
    label_column = "target"

    # if we are asked to use digit names instead of integers
    if use_digit_names:
        # create a dictionary mapping integer labels to fake class names
        label_dict = {
            0: "zero",
            1: "one",
            2: "two",
            3: "three",
            4: "four",
            5: "five",
            6: "six",
            7: "seven",
            8: "eight",
            9: "nine",
        }

        # add a new class column that contains the class names
        df_digits["class"] = df_digits.apply(lambda row: label_dict[row["target"]], axis=1)

        # drop the target column and use "class" as the label column
        df_digits.drop(columns=["target"], axis=1, inplace=True)
        label_column = "class"

    # now subset the training and test rows from the data frame
    df_train = df_digits.iloc[train_indices].copy()
    df_test = df_digits.iloc[test_indices].copy()

    # now create the train and test feature sets from the data frames
    train_fs = FeatureSet.from_data_frame(df_train, "digits_train", labels_column=label_column)
    if len(df_test) > 0:
        test_fs = FeatureSet.from_data_frame(
            df_test, "digits_test", labels_column=label_column, vectorizer=train_fs.vectorizer
        )
    else:
        test_fs = None

    return train_fs, test_fs


def make_california_housing_data(num_examples=None, test_size=0.2):
    """
    Create train/test featuresets from the California housing dataset.

    All columns are standardized into z-scores before creating the
    two feature sets.

    Parameters
    ----------
    num_examples : int, optional
        Number of total examples to use. 80% of these examples
        will be in the training set and the remaining 20% in
        the test set. If ``None``, it will use all of the
        examples in the dataset.
        Defaults to ``None``.
    test_size : float, optional
        Fraction of ``num_examples`` to use for the test
        featureset. Should be between 0 and 1. If this is 0,
        the training featureset will contain all of the
        examples and the test featureset will be ``None``.
        Defaults to 0.2.

    Returns
    -------
    fs_tuple : tuple
        A 2-tuple containing the created train and test featuresets.
        The test featureset will be ``None`` if ``test_size`` is 0.

    Raises
    ------
    ValueError
        If ``num_examples`` is greater than the number of available
        examples.
    """
    # load the housing data
    housing = fetch_california_housing(
        data_home=str(other_dir), download_if_missing=False, as_frame=True
    )
    df_housing = housing.frame

    # standardize all of the values to get them on the same scale
    df_housing = (df_housing - df_housing.mean()) / df_housing.std()

    # use all examples if ``num_examples`` was ``None``
    num_examples = len(df_housing) if num_examples is None else num_examples

    # raise an exception if the number of desired examples is greater
    # than the number of available examples in the data
    if num_examples > len(df_housing):
        raise ValueError(
            f"invalid value {num_examples} for 'num_examples', "
            f"only {len(df_housing)} examples are available."
        )

    # select the desired number of examples
    row_indices = np.arange(len(df_housing))
    prng = np.random.default_rng(123456789)
    if num_examples < len(df_housing):
        chosen_row_indices = prng.choice(
            row_indices, size=num_examples, replace=False, shuffle=False
        )
    else:
        chosen_row_indices = row_indices

    # now split the chosen indices into train and test indices
    # assuming ``test_size`` > 0
    if test_size > 0:
        splitter = ShuffleSplit(n_splits=1, test_size=test_size, random_state=123456789)
        train_indices, test_indices = list(splitter.split(chosen_row_indices))[0]
    else:
        train_indices = chosen_row_indices
        test_indices = np.array([])

    # now subset the training and test rows from the data frame
    df_train = df_housing.iloc[train_indices].copy()
    df_test = df_housing.iloc[test_indices].copy()

    # now create the train and test feature sets from the data frames
    train_fs = FeatureSet.from_data_frame(df_train, "housing_train", labels_column="MedHouseVal")
    if len(df_test) > 0:
        test_fs = FeatureSet.from_data_frame(
            df_test, "housing_test", labels_column="MedHouseVal", vectorizer=train_fs.vectorizer
        )
    else:
        test_fs = None

    return train_fs, test_fs


def compute_expected_folds_for_cv_testing(featureset, num_folds=10, stratified=True, seed=None):
    """
    Compute the fold IDs expected from SKLL's ``cross_validate()`` methods.

    This function is useful for cross-validation tests where
    we are trying to confirm that that SKLL is correctly computing
    the k-fold cross-validation folds for the given data for cases
    where k is a number. If ``seed`` is specified, the function first
    shuffles the data using that seed and then uses a
    ``StratifiedKFold()`` splitter to split the data into
    ``num_folds``.

    Parameters
    ----------
    featureset : skll.data.FeatureSet
        The given featureset for which to compute the folds.
    num_folds : int, optional
        The number of folds into which to split the given featureset.
        Defaults to 10.
    stratified : bool, optional
        Whether to use stratified k-fold splitting or regular splitting.
        Defaults to ``True`` (stratified).
    seed : int, optional
        The seed to use for the random number generator to
        shuffle the data before splitting. If ``None``, no
        shuffling takes place.

    Returns
    -------
    expected_fold_ids : dict
        A dictionary mapping each ID in the featureset to a fold ID.
        Fold IDs range from 0 to ``num_folds``-1.
    """
    # initialize the return dictionary
    expected_fold_ids = {}

    # initialize an RNG with the given seed, if available
    if seed:
        random_state = np.random.RandomState(seed)

        # shuffle and split the featureset ids, labels and features
        ids, labels, features = sk_shuffle(
            featureset.ids, featureset.labels, featureset.features, random_state=random_state
        )

        featureset = FeatureSet(
            featureset.name, ids, labels=labels, features=features, vectorizer=featureset.vectorizer
        )

    # split the featureset IDs into the given number of folds
    # and save the fold ID for each featureset ID
    kfold = StratifiedKFold(n_splits=num_folds) if stratified else KFold(n_splits=num_folds)
    for fold_num, (_, test_indices) in enumerate(
        kfold.split(featureset.features, featureset.labels)
    ):
        for index in test_indices:
            expected_fold_ids[featureset.ids[index]] = str(fold_num)

    return expected_fold_ids



#!/usr/bin/env python
# License: BSD 3 clause
"""
Generate learning plots from the learning curve output TSV file.

This is necessary in scenarios where the plots were not generated as part of
the original learning curve experiment, e.g. the experiment was run on a remote
server where plots may not have been generated either due to a crash or
incorrect setting of the DISPLAY environment variable.

In these cases, the summary file should always be generated and this script can
then be used to generate the plots later.

:author: Nitin Madnani
:organization: ETS
"""

import argparse
import logging
import sys
from pathlib import Path
from typing import List, Optional

from skll.experiments import generate_learning_curve_plots
from skll.version import __version__


def main(argv: Optional[List[str]] = None) -> None:
    """
    Handle command line arguments and gets things started.

    Parameters
    ----------
    argv : Optional[List[str]], default=None
        List of arguments, as if specified on the command-line.
        If ``None``, ``sys.argv[1:]`` is used instead.
    """
    # Get command line arguments
    parser = argparse.ArgumentParser(
        description="Generates learning curve plots from the learning curve " "TSV file.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        conflict_handler="resolve",
    )
    parser.add_argument("tsv_file", help="Learning curve TSV output file.")
    parser.add_argument("output_dir", help="Directory to store the learning curve plots.")
    parser.add_argument("--version", action="version", version=f"%(prog)s {__version__}")
    args = parser.parse_args(argv)

    # Make warnings from built-in warnings module get formatted more nicely
    logging.captureWarnings(True)
    logging.basicConfig(format="%(asctime)s - %(name)s - %(levelname)s - " "%(message)s")

    # convert to Path objects
    tsv_file = Path(args.tsv_file)
    output_dir = Path(args.output_dir)

    # make sure that the input TSV file that's being passed exists
    if not tsv_file.exists():
        logging.error(f"Error: the given file {args.tsv_file} does not " "exist.")
        sys.exit(1)

    # create the output directory if it doesn't already exist
    if not output_dir.exists():
        output_dir.mkdir(parents=True)

    # get the experiment name from the learning curve TSV file
    # output_file_name = experiment_name + '_summary.tsv'
    experiment_name = tsv_file.name.rstrip("_summary.tsv")
    logging.info("Generating learning curve(s)")
    generate_learning_curve_plots(experiment_name, args.output_dir, args.tsv_file)


if __name__ == "__main__":
    main()
